@@ -57,20 +57,22 @@ import sys
 import time
 import logging
 import platform
-from typing import Tuple, Optional, Sequence, List, Any, Iterator, Union, Type
+import itertools
+from typing import Tuple, Optional, Sequence, List, Any, Iterator, Union
 from argparse import ArgumentParser, SUPPRESS, HelpFormatter
 
 import dnaio
 
 from cutadapt import __version__
-from cutadapt.adapters import warn_duplicate_adapters, Adapter
+from cutadapt.adapters import warn_duplicate_adapters, Adapter, InvalidCharacter
 from cutadapt.parser import AdapterParser
-from cutadapt.modifiers import (SingleEndModifier, LengthTagModifier, SuffixRemover, PrefixSuffixAdder,
+from cutadapt.modifiers import (SingleEndModifier, LengthTagModifier, SuffixRemover,
+    PrefixSuffixAdder,
     ZeroCapper, QualityTrimmer, UnconditionalCutter, NEndTrimmer, AdapterCutter,
     PairedAdapterCutterError, PairedAdapterCutter, NextseqQualityTrimmer, Shortener,
-    ReverseComplementer)
-from cutadapt.report import full_report, minimal_report
-from cutadapt.pipeline import (Pipeline, SingleEndPipeline, PairedEndPipeline, InputFiles,
+    ReverseComplementer, PairedEndRenamer, Renamer, InvalidTemplate)
+from cutadapt.report import full_report, minimal_report, Statistics
+from cutadapt.pipeline import (Pipeline, SingleEndPipeline, PairedEndPipeline, InputPaths,
     OutputFiles, PipelineRunner, SerialPipelineRunner, ParallelPipelineRunner)
 from cutadapt.utils import available_cpu_count, Progress, DummyProgress, FileOpener
 from cutadapt.log import setup_logging, REPORT
@@ -85,6 +87,14 @@ class CutadaptArgumentParser(ArgumentParser):
     - A brief message is shown on errors, not full usage
     """
     class CustomUsageHelpFormatter(HelpFormatter):
+        def __init__(self, *args, **kwargs):
+            try:
+                import shutil
+                kwargs['width'] = min(24 + 80, shutil.get_terminal_size().columns)
+            except ImportError:
+                pass
+            super().__init__(*args, **kwargs)
+
         def add_usage(self, usage, actions, groups, prefix=None):
             if usage is not SUPPRESS:
                 args = usage, actions, groups, ''
@@ -116,9 +126,8 @@ def get_argument_parser() -> ArgumentParser:
     group.add_argument("-h", "--help", action="help", help="Show this help message and exit")
     group.add_argument("--version", action="version", help="Show version number and exit",
         version=__version__)
-    group.add_argument("--debug", nargs="?", const=True, default=False,
-        choices=("trace",),
-        help="Print debug log. 'trace' prints also DP matrices")
+    group.add_argument("--debug", action="count", default=0,
+        help="Print debug log. Use twice to also print DP matrices")
     group.add_argument("--profile", action="store_true", default=False, help=SUPPRESS)
     group.add_argument('-j', '--cores', type=int, default=1,
         help='Number of CPU cores to use. Use 0 to auto-detect. Default: %(default)s')
@@ -131,10 +140,10 @@ def get_argument_parser() -> ArgumentParser:
     group.add_argument("--buffer-size", type=int, default=4000000,
         help=SUPPRESS)
     # Compression level for gzipped output files. Not exposed since we have -Z
-    group.add_argument("--compression-level", type=int, default=6,
+    group.add_argument("--compression-level", type=int, default=5,
         help=SUPPRESS)
-    # Deprecated: The input format is always auto-detected
-    group.add_argument("-f", "--format", help=SUPPRESS)
+    # Disable adapter index creation
+    group.add_argument("--no-index", dest="index", default=True, action="store_false", help=SUPPRESS)
 
     group = parser.add_argument_group("Finding adapters",
         description="Parameters -a, -g, -b specify adapters to be removed from "
@@ -160,14 +169,16 @@ def get_argument_parser() -> ArgumentParser:
         default=[], metavar="ADAPTER", dest="adapters",
         help="Sequence of an adapter that may be ligated to the 5' or 3' end "
             "(paired data: of the first read). Both types of matches as "
-            "described under -a und -g are allowed. If the first base of the "
+            "described under -a and -g are allowed. If the first base of the "
             "read is part of the match, the behavior is as with -g, otherwise "
             "as with -a. This option is mostly for rescuing failed library "
             "preparations - do not use if you know which end your adapter was "
             "ligated to!")
-    group.add_argument("-e", "--error-rate", type=float, default=0.1, metavar="RATE",
-        help="Maximum allowed error rate as value between 0 and 1 (no. of "
-            "errors divided by length of matching region). Default: %(default)s (=10%%)")
+    group.add_argument("-e", "--error-rate", "--errors",
+        type=float, metavar="E", default=0.1,
+        help="Maximum allowed error rate (if 0 <= E < 1), or absolute number of errors "
+            "for full-length adapter match (if E is an integer >= 1). Error rate = "
+            "no. of errors divided by length of matching region. Default: %(default)s (10%%)")
     group.add_argument("--no-indels", action='store_false', dest='indels', default=True,
         help="Allow only mismatches in alignments. "
             "Default: allow both mismatches and indels")
@@ -181,12 +192,14 @@ def get_argument_parser() -> ArgumentParser:
     group.add_argument("-N", "--no-match-adapter-wildcards", action="store_false",
         default=True, dest='match_adapter_wildcards',
         help="Do not interpret IUPAC wildcards in adapters.")
-    group.add_argument("--action", choices=('trim', 'mask', 'lowercase', 'none'), default='trim',
-        help="What to do with found adapters. "
+    group.add_argument("--action", choices=("trim", "retain", "mask", "lowercase", "none"),
+        default="trim",
+        help="What to do if a match was found. "
+            "trim: trim adapter and up- or downstream sequence; "
+            "retain: trim, but retain adapter; "
             "mask: replace with 'N' characters; "
             "lowercase: convert to lowercase; "
-            "none: leave unchanged (useful with "
-            "--discard-untrimmed). Default: %(default)s")
+            "none: leave unchanged. Default: %(default)s")
     group.add_argument("--rc", "--revcomp", dest="reverse_complement", default=False,
         action="store_true",
         help="Check both the read and its reverse complement for adapter matches. If "
@@ -235,6 +248,9 @@ def get_argument_parser() -> ArgumentParser:
             "adapter.")
     group.add_argument("-y", "--suffix", default='',
         help="Add this suffix to read names; can also include {name}")
+    group.add_argument("--rename", metavar="TEMPLATE",
+        help="Rename reads using TEMPLATE containing variables such as {id}, {adapter_name} "
+            "etc. (see documentation)")
     group.add_argument("--zero-cap", "-z", action='store_true', default=False,
         help="Change negative quality values to zero.")
 
@@ -327,17 +343,9 @@ def get_argument_parser() -> ArgumentParser:
             "was found. Use with --untrimmed-output. Default: output "
             "to same file as trimmed reads")
     group.add_argument("--too-short-paired-output", metavar="FILE", default=None,
-        help="Write second read in a pair to this file if pair is too short. "
-            "Use also --too-short-output.")
+        help="Write second read in a pair to this file if pair is too short.")
     group.add_argument("--too-long-paired-output", metavar="FILE", default=None,
-        help="Write second read in a pair to this file if pair is too long. "
-            "Use also --too-long-output.")
-
-    for arg in ("--colorspace", "-c", "-d", "--double-encode", "-t", "--trim-primer",
-            "--strip-f3", "--maq", "--bwa", "--no-zero-cap"):
-        group.add_argument(arg, dest='colorspace', action='store_true', default=False,
-        help=SUPPRESS)
-    parser.set_defaults(colorspace=False)
+        help="Write second read in a pair to this file if pair is too long.")
 
     # We could have two positional arguments here, with the second one optional, but
     # we want custom, more helpful error messages.
@@ -393,7 +401,7 @@ def parse_lengths(s: str) -> Tuple[Optional[int], ...]:
 
 
 def open_output_files(
-    args, default_outfile, file_opener: FileOpener
+    args, default_outfile, file_opener: FileOpener, adapter_names: Sequence[str], adapter_names2: Sequence[str],
 ) -> OutputFiles:
     """
     Return an OutputFiles instance. If demultiplex is True, the untrimmed, untrimmed2, out and out2
@@ -424,35 +432,20 @@ def open_output_files(
         raise CommandLineError("Do not use --discard-trimmed when demultiplexing.")
 
     if demultiplex_mode == "normal":
-        out = args.output
-        untrimmed = args.output.replace('{name}', 'unknown')
-        if args.untrimmed_output:
-            untrimmed = args.untrimmed_output
-        if args.discard_untrimmed:
-            untrimmed = None
-
-        if args.paired_output is not None:
-            out2 = args.paired_output
-            untrimmed2 = args.paired_output.replace('{name}', 'unknown')
-            if args.untrimmed_paired_output:
-                untrimmed2 = args.untrimmed_paired_output
-            if args.discard_untrimmed:
-                untrimmed2 = None
-        else:
-            untrimmed2 = out2 = None
-
-        assert out is not None and '{name}' in out and (out2 is None or '{name}' in out2)
+        out = out2 = None
+        combinatorial_out = combinatorial_out2 = None
+        demultiplex_out, demultiplex_out2, untrimmed, untrimmed2 = open_demultiplex_out(
+            adapter_names, args, file_opener)
     elif demultiplex_mode == "combinatorial":
-        out = args.output
-        out2 = args.paired_output
-        if args.untrimmed_output or args.untrimmed_paired_output:
-            raise CommandLineError("Combinatorial demultiplexing (with {name1} and {name2})"
-                " cannot be combined with --untrimmed-output or --untrimmed-paired-output")
-        if args.discard_untrimmed:
-            untrimmed = untrimmed2 = None
-        else:
-            untrimmed = untrimmed2 = 'unknown'
+        assert '{name1}' in args.output and '{name2}' in args.output
+        assert '{name1}' in args.paired_output and '{name2}' in args.paired_output
+        out = out2 = None
+        demultiplex_out = demultiplex_out2 = None
+        combinatorial_out, combinatorial_out2, untrimmed, untrimmed2 = open_combinatorial_out(
+            adapter_names, adapter_names2, args, file_opener)
     else:
+        combinatorial_out = combinatorial_out2 = None
+        demultiplex_out = demultiplex_out2 = None
         untrimmed, untrimmed2 = file_opener.xopen_pair(
             args.untrimmed_output, args.untrimmed_paired_output, "wb")
         out, out2 = file_opener.xopen_pair(args.output, args.paired_output, "wb")
@@ -471,11 +464,67 @@ def open_output_files(
         untrimmed2=untrimmed2,
         out=out,
         out2=out2,
-        demultiplex=bool(demultiplex_mode),
+        demultiplex_out=demultiplex_out,
+        demultiplex_out2=demultiplex_out2,
+        combinatorial_out=combinatorial_out,
+        combinatorial_out2=combinatorial_out2,
         force_fasta=args.fasta,
     )
 
 
+def open_combinatorial_out(adapter_names, adapter_names2, args, file_opener):
+    combinatorial_out = dict()
+    combinatorial_out2 = dict()
+    if args.discard_untrimmed:
+        extra = []
+    else:
+        extra = [(None, None)]
+        extra += [(None, name2) for name2 in adapter_names2]
+        extra += [(name1, None) for name1 in adapter_names]
+    for name1, name2 in list(itertools.product(adapter_names, adapter_names2)) + extra:
+        fname1 = name1 if name1 is not None else "unknown"
+        fname2 = name2 if name2 is not None else "unknown"
+        path1 = args.output.replace("{name1}", fname1).replace("{name2}", fname2)
+        path2 = args.paired_output.replace("{name1}", fname1).replace("{name2}", fname2)
+        combinatorial_out[(name1, name2)] = file_opener.xopen(path1, "wb")
+        combinatorial_out2[(name1, name2)] = file_opener.xopen(path2, "wb")
+    if args.untrimmed_output or args.untrimmed_paired_output:
+        raise CommandLineError(
+            "Combinatorial demultiplexing (with {name1} and {name2})"
+            " cannot be combined with --untrimmed-output or --untrimmed-paired-output")
+    untrimmed = untrimmed2 = None
+    return combinatorial_out, combinatorial_out2, untrimmed, untrimmed2
+
+
+def open_demultiplex_out(adapter_names, args, file_opener):
+    demultiplex_out = dict()
+    demultiplex_out2 = dict() if args.paired_output is not None else None
+    for name in adapter_names:
+        path1 = args.output.replace("{name}", name)
+        demultiplex_out[name] = file_opener.xopen(path1, "wb")
+        if args.paired_output is not None:
+            path2 = args.paired_output.replace("{name}", name)
+            demultiplex_out2[name] = file_opener.xopen(path2, "wb")
+    untrimmed_path = args.output.replace("{name}", "unknown")
+    if args.untrimmed_output:
+        untrimmed_path = args.untrimmed_output
+    if args.discard_untrimmed:
+        untrimmed = None
+    else:
+        untrimmed = file_opener.xopen(untrimmed_path, "wb")
+    if args.paired_output is not None:
+        untrimmed2_path = args.paired_output.replace("{name}", "unknown")
+        if args.untrimmed_paired_output:
+            untrimmed2_path = args.untrimmed_paired_output
+        if args.discard_untrimmed:
+            untrimmed2 = None
+        else:
+            untrimmed2 = file_opener.xopen(untrimmed2_path, "wb")
+    else:
+        untrimmed2 = None
+    return demultiplex_out, demultiplex_out2, untrimmed, untrimmed2
+
+
 def determine_demultiplex_mode(args) -> Union[str, bool]:
     """Return one of "normal", "combinatorial" or False"""
 
@@ -585,12 +634,6 @@ def check_arguments(args, paired: bool) -> None:
                     " --{name}-output/--{name}-paired-output options.".format(name=argname)
                 )
 
-    if args.format is not None:
-        logger.warning("Option --format is deprecated and ignored because the input file format is "
-            "always auto-detected")
-
-    if not (0 <= args.error_rate < 1.):
-        raise CommandLineError("The maximum error rate must be at least 0 and less than 1.")
     if args.overlap < 1:
         raise CommandLineError("The overlap must be at least 1.")
     if not (0 <= args.gc_content <= 100):
@@ -600,7 +643,7 @@ def check_arguments(args, paired: bool) -> None:
         raise CommandLineError("--pair-adapters cannot be used with --times")
 
 
-def pipeline_from_parsed_args(args, paired, file_opener) -> Pipeline:
+def pipeline_from_parsed_args(args, paired, file_opener, adapters, adapters2) -> Pipeline:
     """
     Setup a processing pipeline from parsed command-line arguments.
 
@@ -611,8 +654,6 @@ def pipeline_from_parsed_args(args, paired, file_opener) -> Pipeline:
     if args.action == 'none':
         args.action = None
 
-    adapters, adapters2 = adapters_from_args(args)
-
     # Create the processing pipeline
     if paired:
         pair_filter_mode = 'any' if args.pair_filter is None else args.pair_filter
@@ -647,11 +688,24 @@ def pipeline_from_parsed_args(args, paired, file_opener) -> Pipeline:
         args.action,
         args.times,
         args.reverse_complement,
+        args.index,
     )
 
     for modifier in modifiers_applying_to_both_ends_if_paired(args):
         pipeline_add(modifier)
 
+    if args.rename:
+        if args.prefix or args.suffix:
+            raise CommandLineError(
+                "Option --rename cannot be combined with --prefix (-x) or --suffix (-y)"
+            )
+        try:
+            if paired:
+                pipeline.add_paired_modifier(PairedEndRenamer(args.rename))
+            else:
+                pipeline_add(Renamer(args.rename))
+        except InvalidTemplate as e:
+            raise CommandLineError(e)
     # Set filtering parameters
     # Minimum/maximum length
     for attr in 'minimum_length', 'maximum_length':
@@ -674,7 +728,7 @@ def pipeline_from_parsed_args(args, paired, file_opener) -> Pipeline:
 
 def adapters_from_args(args) -> Tuple[List[Adapter], List[Adapter]]:
     adapter_parser = AdapterParser(
-        max_error_rate=args.error_rate,
+        max_errors=args.error_rate,
         min_overlap=args.overlap,
         read_wildcards=args.match_read_wildcards,
         adapter_wildcards=args.match_adapter_wildcards,
@@ -683,11 +737,11 @@ def adapters_from_args(args) -> Tuple[List[Adapter], List[Adapter]]:
     try:
         adapters = adapter_parser.parse_multi(args.adapters)
         adapters2 = adapter_parser.parse_multi(args.adapters2)
-    except (FileNotFoundError, ValueError) as e:
+    except (FileNotFoundError, ValueError, InvalidCharacter) as e:
         raise CommandLineError(e)
     warn_duplicate_adapters(adapters)
     warn_duplicate_adapters(adapters2)
-    if args.debug == "trace":
+    if args.debug > 1:
         for adapter in adapters + adapters2:
             adapter.enable_debug()
     return adapters, adapters2
@@ -724,9 +778,10 @@ def add_adapter_cutter(
     adapters2,
     paired: bool,
     pair_adapters: bool,
-    action: str,
+    action: Optional[str],
     times: int,
     reverse_complement: bool,
+    allow_index: bool,
 ):
     if pair_adapters:
         if reverse_complement:
@@ -738,10 +793,13 @@ def add_adapter_cutter(
         pipeline.add_paired_modifier(cutter)
     else:
         adapter_cutter, adapter_cutter2 = None, None
-        if adapters:
-            adapter_cutter = AdapterCutter(adapters, times, action)
-        if adapters2:
-            adapter_cutter2 = AdapterCutter(adapters2, times, action)
+        try:
+            if adapters:
+                adapter_cutter = AdapterCutter(adapters, times, action, allow_index)
+            if adapters2:
+                adapter_cutter2 = AdapterCutter(adapters2, times, action, allow_index)
+        except ValueError as e:
+            raise CommandLineError(e)
         if paired:
             if reverse_complement:
                 raise CommandLineError("--revcomp not implemented for paired-end reads")
@@ -782,42 +840,36 @@ def log_header(cmdlineargs):
     logger.info("Command line parameters: %s", " ".join(cmdlineargs))
 
 
-def main(cmdlineargs=None, default_outfile=sys.stdout.buffer):
+def main_cli():  # pragma: no cover
+    """Entry point for command-line script"""
+    main(sys.argv[1:])
+    return 0
+
+
+def main(cmdlineargs, default_outfile=sys.stdout.buffer) -> Statistics:
     """
-    Main function that sets up a processing pipeline and runs it.
+    Set up a processing pipeline from the command-line arguments, run it and return
+    a Statistics object.
 
     default_outfile is the file to which trimmed reads are sent if the ``-o``
     parameter is not used.
     """
     start_time = time.time()
     parser = get_argument_parser()
-    if cmdlineargs is None:
-        cmdlineargs = sys.argv[1:]
     args, leftover_args = parser.parse_known_args(args=cmdlineargs)
-    # log to stderr if results are to be sent to stdout
-    log_to_stdout = args.output is not None and args.output != "-" and args.paired_output != "-"
     # Setup logging only if there are not already any handlers (can happen when
     # this function is being called externally such as from unit tests)
     if not logging.root.handlers:
+        # If results are to be sent to stdout, logging needs to go to stderr
+        log_to_stdout = args.output is not None and args.output != "-" and args.paired_output != "-"
         setup_logging(logger, stdout=log_to_stdout,
             quiet=args.quiet, minimal=args.report == 'minimal', debug=args.debug)
+    log_header(cmdlineargs)
     profiler = setup_profiler_if_requested(args.profile)
 
     if args.quiet and args.report:
         parser.error("Options --quiet and --report cannot be used at the same time")
 
-    if args.colorspace:
-        parser.error(
-            "These colorspace-specific options are no longer supported: "
-            "--colorspace, -c, -d, --double-encode, -t, --trim-primer, "
-            "--strip-f3, --maq, --bwa, --no-zero-cap. "
-            "Use Cutadapt 1.18 or earlier to work with colorspace data.")
-
-    paired = determine_paired(args)
-    assert paired in (False, True)
-
-    # Print the header now because some of the functions below create logging output
-    log_header(cmdlineargs)
     if leftover_args:
         warn_if_en_dashes(cmdlineargs)
         parser.error("unrecognized arguments: " + " ".join(leftover_args))
@@ -832,20 +884,25 @@ def main(cmdlineargs=None, default_outfile=sys.stdout.buffer):
         progress = Progress()
     else:
         progress = DummyProgress()
+    paired = determine_paired(args)
+    assert paired in (False, True)
 
     try:
         is_interleaved_input = args.interleaved and len(args.inputs) == 1
         input_filename, input_paired_filename = setup_input_files(args.inputs,
             paired, is_interleaved_input)
         check_arguments(args, paired)
-        pipeline = pipeline_from_parsed_args(args, paired, file_opener)
-        outfiles = open_output_files(args, default_outfile, file_opener)
-        infiles = InputFiles(input_filename, file2=input_paired_filename,
-            interleaved=is_interleaved_input)
-        runner = setup_runner(pipeline, infiles, outfiles, progress, cores, args.buffer_size)
+        adapters, adapters2 = adapters_from_args(args)
+        pipeline = pipeline_from_parsed_args(args, paired, file_opener, adapters, adapters2)
+        adapter_names = [a.name for a in adapters]  # type: List[str]
+        adapter_names2 = [a.name for a in adapters2]  # type: List[str]
+        outfiles = open_output_files(args, default_outfile, file_opener, adapter_names, adapter_names2)
+        inpaths = InputPaths(input_filename, path2=input_paired_filename, interleaved=is_interleaved_input)
+        runner = setup_runner(pipeline, inpaths, outfiles, progress, cores, args.buffer_size, file_opener)
     except CommandLineError as e:
+        logger.debug("Command line error. Traceback:", exc_info=True)
         parser.error(str(e))
-        return  # avoid IDE warnings below
+        return
 
     logger.info("Processing reads on %d core%s in %s mode ...",
         cores, 's' if cores > 1 else '',
@@ -859,6 +916,7 @@ def main(cmdlineargs=None, default_outfile=sys.stdout.buffer):
     except BrokenPipeError:
         sys.exit(1)
     except (dnaio.FileFormatError, dnaio.UnknownFileFormat, EOFError) as e:
+        logger.debug("Command line error. Traceback:", exc_info=True)
         sys.exit("cutadapt: error: {}".format(e))
 
     elapsed = time.time() - start_time
@@ -871,28 +929,28 @@ def main(cmdlineargs=None, default_outfile=sys.stdout.buffer):
         import pstats
         profiler.disable()
         pstats.Stats(profiler).sort_stats('time').print_stats(20)
+    return stats
 
 
-def setup_runner(pipeline: Pipeline, infiles, outfiles, progress, cores, buffer_size):
-    if cores > 1:
-        if ParallelPipelineRunner.can_output_to(outfiles):
-            runner_class = ParallelPipelineRunner  # type: Type[PipelineRunner]
-            runner_kwargs = dict(n_workers=cores, buffer_size=buffer_size)
-        else:
-            raise CommandLineError("Running in parallel is currently not supported "
-                 "when using --format or when demultiplexing.\n"
-                 "Omit --cores/-j to continue.")
-            # return  # avoid IDE warnings below
-    else:
-        runner_class = SerialPipelineRunner
-        runner_kwargs = dict()
+def setup_runner(
+    pipeline: Pipeline,
+    inpaths: InputPaths,
+    outfiles: OutputFiles,
+    progress: Progress,
+    cores: int,
+    buffer_size: int,
+    file_opener: FileOpener,
+) -> PipelineRunner:
     try:
-        runner = runner_class(pipeline, infiles, outfiles, progress, **runner_kwargs)
-    except (dnaio.UnknownFileFormat, OSError) as e:
+        if cores > 1:
+            return ParallelPipelineRunner(
+                pipeline, inpaths, outfiles, progress, n_workers=cores, buffer_size=buffer_size)
+        else:
+            infiles = inpaths.open(file_opener)
+            return SerialPipelineRunner(pipeline, infiles, outfiles, progress)
+    except (dnaio.UnknownFileFormat, dnaio.FileFormatError, OSError) as e:
         raise CommandLineError(e)
 
-    return runner
-
 
 def setup_profiler_if_requested(requested):
     if requested:
@@ -914,5 +972,5 @@ def warn_if_en_dashes(args):
             )
 
 
-if __name__ == '__main__':
-    main()
+if __name__ == '__main__':  # pragma: no cover
+    sys.exit(main_cli())
