@@ -15,6 +15,10 @@ from . import align
 logger = logging.getLogger()
 
 
+class InvalidCharacter(Exception):
+    pass
+
+
 class Where(Enum):
     # Constants for the Aligner.locate() function.
     # The function is called with SEQ1 as the adapter, SEQ2 as the read.
@@ -143,10 +147,16 @@ class AdapterStatistics:
 
 class Match(ABC):
 
+    adapter: "Adapter"
+
     @abstractmethod
     def remainder_interval(self) -> Tuple[int, int]:
         pass
 
+    @abstractmethod
+    def retained_adapter_interval(self) -> Tuple[int, int]:
+        pass
+
     @abstractmethod
     def get_info_records(self, read) -> List[List]:
         pass
@@ -192,6 +202,19 @@ class SingleMatch(Match, ABC):
         return 'SingleMatch(astart={}, astop={}, rstart={}, rstop={}, matches={}, errors={})'.format(
             self.astart, self.astop, self.rstart, self.rstop, self.matches, self.errors)
 
+    def __eq__(self, other) -> bool:
+        return (
+            other.__class__ is self.__class__
+            and self.astart == other.astart
+            and self.astop == other.astop
+            and self.rstart == other.rstart
+            and self.rstop == other.rstop
+            and self.matches == other.matches
+            and self.errors == other.errors
+            and self.adapter is other.adapter
+            and self.sequence == other.sequence
+        )
+
     def wildcards(self, wildcard_char: str = "N") -> str:
         """
         Return a string that contains, for each wildcard character,
@@ -254,6 +277,9 @@ class RemoveBeforeMatch(SingleMatch):
         """
         return self.rstop, len(self.sequence)
 
+    def retained_adapter_interval(self) -> Tuple[int, int]:
+        return self.rstart, len(self.sequence)
+
     def trim_slice(self):
         # Same as remainder_interval, but as a slice() object
         return slice(self.rstop, None)
@@ -289,6 +315,9 @@ class RemoveAfterMatch(SingleMatch):
         """
         return 0, self.rstart
 
+    def retained_adapter_interval(self) -> Tuple[int, int]:
+        return 0, self.rstop
+
     def trim_slice(self):
         # Same as remainder_interval, but as a slice() object
         return slice(None, self.rstart)
@@ -312,11 +341,10 @@ def _generate_adapter_name(_start=[1]) -> str:
     return name
 
 
-class Adapter(ABC):
+class Matchable(ABC):
+    """Something that has a match_to() method."""
 
-    description = "adapter with one component"  # this is overriden in subclasses
-
-    def __init__(self, name: Optional[str], *args, **kwargs):
+    def __init__(self, name: str, *args, **kwargs):
         self.name = name
 
     @abstractmethod
@@ -328,6 +356,15 @@ class Adapter(ABC):
         pass
 
 
+class Adapter(Matchable, ABC):
+
+    description = "adapter with one component"  # this is overriden in subclasses
+
+    @abstractmethod
+    def create_statistics(self) -> AdapterStatistics:
+        pass
+
+
 class SingleAdapter(Adapter, ABC):
     """
     This class can find a single adapter characterized by sequence, error rate,
@@ -339,8 +376,11 @@ class SingleAdapter(Adapter, ABC):
     sequence -- The adapter sequence as string. Will be converted to uppercase.
         Also, Us will be converted to Ts.
 
-    max_error_rate -- Maximum allowed error rate. The error rate is
-        the number of errors in the alignment divided by the length
+    max_errors -- Maximum allowed errors (non-negative float). If the values is less than 1, this is
+        interpreted as a rate directly and passed to the aligner. If it is 1 or greater, the value
+        is converted to a rate by dividing it by the length of the sequence.
+
+        The error rate is the number of errors in the alignment divided by the length
         of the part of the alignment that matches the adapter.
 
     minimum_overlap -- Minimum length of the part of the alignment
@@ -358,35 +398,37 @@ class SingleAdapter(Adapter, ABC):
     def __init__(
         self,
         sequence: str,
-        max_error_rate: float = 0.1,
+        max_errors: float = 0.1,
         min_overlap: int = 3,
         read_wildcards: bool = False,
         adapter_wildcards: bool = True,
         name: Optional[str] = None,
         indels: bool = True,
     ):
-        super().__init__(name)
-        self._debug = False  # type: bool
         self.name = _generate_adapter_name() if name is None else name  # type: str
+        super().__init__(self.name)
+        self._debug = False  # type: bool
         self.sequence = sequence.upper().replace("U", "T")  # type: str
         if not self.sequence:
             raise ValueError("Adapter sequence is empty")
-        self.max_error_rate = max_error_rate  # type: float
+        if max_errors >= 1:
+            max_errors /= len(self.sequence)
+        self.max_error_rate = max_errors  # type: float
         self.min_overlap = min(min_overlap, len(self.sequence))  # type: int
-        iupac = frozenset('XACGTURYSWKMBDHVN')
+        iupac = frozenset('ABCDGHKMNRSTUVWXY')
         if adapter_wildcards and not set(self.sequence) <= iupac:
             for c in self.sequence:
                 if c not in iupac:
-                    raise ValueError('Character {!r} in adapter sequence {!r} is '
+                    raise InvalidCharacter('Character {!r} in adapter sequence {!r} is '
                         'not a valid IUPAC code. Use only characters '
-                        'XACGTURYSWKMBDHVN.'.format(c, self.sequence))
+                        'ABCDGHKMNRSTUVWXY.'.format(c, self.sequence))
         # Optimization: Use non-wildcard matching if only ACGT is used
         self.adapter_wildcards = adapter_wildcards and not set(self.sequence) <= set("ACGT")  # type: bool
         self.read_wildcards = read_wildcards  # type: bool
         self.indels = indels  # type: bool
         self.aligner = self._aligner()
 
-    def _make_aligner(self, flags):
+    def _make_aligner(self, flags: int) -> align.Aligner:
         # TODO
         # Indels are suppressed by setting their cost very high, but a different algorithm
         # should be used instead.
@@ -409,15 +451,10 @@ class SingleAdapter(Adapter, ABC):
             'indels={indels})>'.format(cls=self.__class__.__name__, **vars(self))
 
     @property
-    def is_anchored(self):
-        """Return whether this adapter is anchored"""
-        return self.where in {Where.PREFIX, Where.SUFFIX}
-
-    @property
-    def effective_length(self):
+    def effective_length(self) -> int:
         return self.aligner.effective_length
 
-    def enable_debug(self):
+    def enable_debug(self) -> None:
         """
         Print out the dynamic programming matrix after matching a read to an
         adapter.
@@ -439,10 +476,10 @@ class SingleAdapter(Adapter, ABC):
         overlap length, maximum error rate).
         """
 
-    def __len__(self):
+    def __len__(self) -> int:
         return len(self.sequence)
 
-    def create_statistics(self):
+    def create_statistics(self) -> AdapterStatistics:
         return AdapterStatistics(self, self)
 
 
@@ -455,7 +492,7 @@ class FrontAdapter(SingleAdapter):
         self._force_anywhere = kwargs.pop("force_anywhere", False)
         super().__init__(*args, **kwargs)
 
-    def _aligner(self):
+    def _aligner(self) -> align.Aligner:
         return self._make_aligner(Where.ANYWHERE.value if self._force_anywhere else Where.FRONT.value)
 
     def match_to(self, sequence: str):
@@ -466,9 +503,9 @@ class FrontAdapter(SingleAdapter):
         return None if no match was found given the matching criteria (minimum
         overlap length, maximum error rate).
         """
-        alignment = self.aligner.locate(sequence.upper())  # type: Optional[Tuple[int,int,int,int,int,int]]
+        alignment = self.aligner.locate(sequence)  # type: Optional[Tuple[int,int,int,int,int,int]]
         if self._debug:
-            print(self.aligner.dpmatrix)  # pragma: no cover
+            print(self.aligner.dpmatrix)
         if alignment is None:
             return None
         return RemoveBeforeMatch(*alignment, adapter=self, sequence=sequence)
@@ -494,7 +531,7 @@ class BackAdapter(SingleAdapter):
         return None if no match was found given the matching criteria (minimum
         overlap length, maximum error rate).
         """
-        alignment = self.aligner.locate(sequence.upper())  # type: Optional[Tuple[int,int,int,int,int,int]]
+        alignment = self.aligner.locate(sequence)  # type: Optional[Tuple[int,int,int,int,int,int]]
         if self._debug:
             print(self.aligner.dpmatrix)  # pragma: no cover
         if alignment is None:
@@ -521,9 +558,9 @@ class AnywhereAdapter(SingleAdapter):
         return None if no match was found given the matching criteria (minimum
         overlap length, maximum error rate).
         """
-        alignment = self.aligner.locate(sequence)
+        alignment = self.aligner.locate(sequence.upper())
         if self._debug:
-            print(self.aligner.dpmatrix)  # pragma: no cover
+            print(self.aligner.dpmatrix)
         if alignment is None:
             return None
         # guess: if alignment starts at pos 0, it’s a 5' adapter
@@ -543,15 +580,11 @@ class NonInternalFrontAdapter(FrontAdapter):
         return self._make_aligner(Where.FRONT_NOT_INTERNAL.value)
 
     def match_to(self, sequence: str):
-        if not self.adapter_wildcards and sequence.startswith(self.sequence):
-            n = len(self.sequence)
-            return RemoveBeforeMatch(
-                0, n, 0, n, n, 0, adapter=self, sequence=sequence
-            )  # type: ignore
+        # The locate function takes care of uppercasing the sequence
         alignment = self.aligner.locate(sequence)
         if self._debug:
             try:
-                print(self.aligner.dpmatrix)  # pragma: no cover
+                print(self.aligner.dpmatrix)
             except AttributeError:
                 pass
         if alignment is None:
@@ -568,13 +601,7 @@ class NonInternalBackAdapter(BackAdapter):
         return self._make_aligner(Where.BACK_NOT_INTERNAL.value)
 
     def match_to(self, sequence: str):
-        if not self.adapter_wildcards and sequence.endswith(self.sequence):
-            # Exact match found
-            # astart, astop, rstart, rstop, matches, errors
-            n = len(self.sequence)
-            return RemoveAfterMatch(
-                0, n, len(sequence) - n, len(sequence), n, 0, adapter=self, sequence=sequence
-            )  # type: ignore
+        # The locate function takes care of uppercasing the sequence
         alignment = self.aligner.locate(sequence)
         if self._debug:
             try:
@@ -678,6 +705,18 @@ class LinkedMatch(Match):
         matches = [match for match in [self.front_match, self.back_match] if match is not None]
         return remainder(matches)
 
+    def retained_adapter_interval(self) -> Tuple[int, int]:
+        if self.front_match:
+            start = self.front_match.rstart
+            offset = self.front_match.rstop
+        else:
+            start = offset = 0
+        if self.back_match:
+            end = self.back_match.rstop + offset
+        else:
+            end = len(self.front_match.sequence)
+        return start, end
+
     def get_info_records(self, read) -> List[List]:
         records = []
         for match, namesuffix in [
@@ -700,11 +739,11 @@ class LinkedAdapter(Adapter):
 
     def __init__(
         self,
-        front_adapter,
-        back_adapter,
-        front_required,
-        back_required,
-        name,
+        front_adapter: SingleAdapter,
+        back_adapter: SingleAdapter,
+        front_required: bool,
+        back_required: bool,
+        name: str,
     ):
         super().__init__(name)
         self.front_required = front_required
@@ -735,7 +774,7 @@ class LinkedAdapter(Adapter):
             return None
         return LinkedMatch(front_match, back_match, self)
 
-    def create_statistics(self):
+    def create_statistics(self) -> AdapterStatistics:
         return AdapterStatistics(self, self.front_adapter, self.back_adapter)
 
     @property
@@ -747,35 +786,82 @@ class LinkedAdapter(Adapter):
         return None
 
 
-class MultiAdapter(Adapter, ABC):
+class MultipleAdapters(Matchable):
+    """
+    Represent multiple adapters at once
+    """
+    def __init__(self, adapters: Sequence[Matchable]):
+        super().__init__(name="multiple_adapters")
+        self._adapters = adapters
+
+    def enable_debug(self):
+        for a in self._adapters:
+            a.enable_debug()
+
+    def __getitem__(self, item):
+        return self._adapters[item]
+
+    def __len__(self):
+        return len(self._adapters)
+
+    def match_to(self, sequence: str) -> Optional[SingleMatch]:
+        """
+        Find the adapter that best matches the sequence.
+
+        Return either a Match instance or None if there are no matches.
+        """
+        best_match = None
+        for adapter in self._adapters:
+            match = adapter.match_to(sequence)
+            if match is None:
+                continue
+
+            # the no. of matches determines which adapter fits best
+            if best_match is None or match.matches > best_match.matches or (
+                match.matches == best_match.matches and match.errors < best_match.errors
+            ):
+                best_match = match
+        return best_match
+
+
+class IndexedAdapters(Matchable, ABC):
     """
     Represent multiple adapters of the same type at once and use an index data structure
     to speed up matching. This acts like a "normal" Adapter as it provides a match_to
     method, but is faster with lots of adapters.
 
     There are quite a few restrictions:
-    - no indels are allowed
     - the error rate allows at most 2 mismatches
     - wildcards in the adapter are not allowed
     - wildcards in the read are not allowed
 
     Use the is_acceptable() method to check individual adapters.
     """
-    MultiAdapterIndex = Dict[str, Tuple[SingleAdapter, int, int]]
+    AdapterIndex = Dict[str, Tuple[SingleAdapter, int, int]]
 
     def __init__(self, adapters):
         """All given adapters must be of the same type"""
-        super().__init__(name="multi_adapter")
+        super().__init__(name="indexed_adapters")
         if not adapters:
             raise ValueError("Adapter list is empty")
         for adapter in adapters:
             self._accept(adapter)
         self._adapters = adapters
+        self._multiple_adapters = MultipleAdapters(adapters)
         self._lengths, self._index = self._make_index()
+        logger.debug("String lengths in the index: %s", sorted(self._lengths, reverse=True))
+        if len(self._lengths) == 1:
+            self._length = self._lengths[0]
+            self.match_to = self._match_to_one_length
+        else:
+            self.match_to = self._match_to_multiple_lengths
         self._make_affix = self._get_make_affix()
 
     def __repr__(self):
-        return "MultiAdapter(adapters={!r})".format(self._adapters)
+        return "{}(adapters={!r})".format(self.__class__.__name__, self._adapters)
+
+    def match_to(self, sequence: str):
+        """Never called because it gets overwritten in __init__"""
 
     @abstractmethod
     def _get_make_affix(self):
@@ -793,15 +879,13 @@ class MultiAdapter(Adapter, ABC):
         if adapter.adapter_wildcards:
             raise ValueError("Wildcards in the adapter not supported")
         k = int(len(adapter) * adapter.max_error_rate)
-        if k > 0 and adapter.indels:
-            raise ValueError("Indels not allowed")
         if k > 2:
             raise ValueError("Error rate too high")
 
     @classmethod
     def is_acceptable(cls, adapter):
         """
-        Return whether this adapter is acceptable for being used by MultiAdapter
+        Return whether this adapter is acceptable for being used in an index
 
         Adapters are not acceptable if they allow wildcards, allow too many errors,
         or would lead to a very large index.
@@ -812,15 +896,16 @@ class MultiAdapter(Adapter, ABC):
             return False
         return True
 
-    def _make_index(self) -> Tuple[List[int], "MultiAdapterIndex"]:
+    def _make_index(self) -> Tuple[List[int], "AdapterIndex"]:
         logger.info('Building index of %s adapters ...', len(self._adapters))
-        index = dict()  # type: MultiAdapter.MultiAdapterIndex
+        index = dict()  # type: IndexedAdapters.AdapterIndex
         lengths = set()
         has_warned = False
         for adapter in self._adapters:
             sequence = adapter.sequence
             k = int(adapter.max_error_rate * len(sequence))
-            for s, errors, matches in align.hamming_environment(sequence, k):
+            environment = align.edit_environment if adapter.indels else align.hamming_environment
+            for s, errors, matches in environment(sequence, k):
                 if s in index:
                     other_adapter, other_errors, other_matches = index[s]
                     if matches < other_matches:
@@ -841,27 +926,50 @@ class MultiAdapter(Adapter, ABC):
 
         return sorted(lengths, reverse=True), index
 
-    def match_to(self, sequence: str):
+    def _match_to_one_length(self, sequence: str):
         """
         Match the adapters against a string and return a Match that represents
         the best match or None if no match was found
         """
+        affix = self._make_affix(sequence.upper(), self._length)
+        if "N" in affix:
+            # Fall back to non-indexed matching
+            return self._multiple_adapters.match_to(sequence)
+        try:
+            adapter, e, m = self._index[affix]
+        except KeyError:
+            return None
+        return self._make_match(adapter, self._length, m, e, sequence)
+
+    def _match_to_multiple_lengths(self, sequence: str):
+        """
+        Match the adapters against a string and return a Match that represents
+        the best match or None if no match was found
+        """
+        affix = sequence.upper()
+
         # Check all the prefixes or suffixes (affixes) that could match
         best_adapter = None  # type: Optional[SingleAdapter]
         best_length = 0
         best_m = -1
         best_e = 1000
+        check_n = True
         for length in self._lengths:
             if length < best_m:
                 # No chance of getting the same or a higher number of matches, so we can stop early
                 break
-
-            affix = self._make_affix(sequence, length)
+            affix = self._make_affix(affix, length)
+            if check_n:
+                if "N" in affix:
+                    return self._multiple_adapters.match_to(sequence)
+                check_n = False
             try:
                 adapter, e, m = self._index[affix]
             except KeyError:
                 continue
             if m > best_m or (m == best_m and e < best_e):
+                # TODO this could be made to work:
+                # assert best_m == -1
                 best_adapter = adapter
                 best_e = e
                 best_m = m
@@ -870,14 +978,13 @@ class MultiAdapter(Adapter, ABC):
         if best_m == -1:
             return None
         else:
-            assert best_adapter is not None
             return self._make_match(best_adapter, best_length, best_m, best_e, sequence)
 
     def enable_debug(self):
         pass
 
 
-class MultiPrefixAdapter(MultiAdapter):
+class IndexedPrefixAdapters(IndexedAdapters):
 
     @classmethod
     def _accept(cls, adapter):
@@ -905,7 +1012,7 @@ class MultiPrefixAdapter(MultiAdapter):
         return s[:n]
 
 
-class MultiSuffixAdapter(MultiAdapter):
+class IndexedSuffixAdapters(IndexedAdapters):
 
     @classmethod
     def _accept(cls, adapter):
