@@ -4,26 +4,38 @@ A modifier must be callable and typically implemented as a class with a
 __call__ method.
 """
 import re
-from typing import Sequence, List, Optional
+from types import SimpleNamespace
+from typing import Sequence, List, Tuple, Optional, Set
 from abc import ABC, abstractmethod
 from collections import OrderedDict
 
+from dnaio import record_names_match, Sequence as DnaSequence
+
 from .qualtrim import quality_trim_index, nextseq_trim_index
-from .adapters import MultiPrefixAdapter, MultiSuffixAdapter, Match, remainder
+from .adapters import MultipleAdapters, SingleAdapter, IndexedPrefixAdapters, IndexedSuffixAdapters, \
+    Match, remainder, Adapter
+from .tokenizer import tokenize_braces, TokenizeError, Token, BraceToken
 from .utils import reverse_complemented_sequence
 
 
+# If the number of prefix or suffix adapters is higher than this, switch to using an index
+INDEXING_THRESHOLD = 5
+
+
 class ModificationInfo:
     """
     An object of this class is created for each read that passes through the pipeline.
     Any information (except the read itself) that needs to be passed from one modifier
     to one later in the pipeline or from one modifier to the filters is recorded here.
     """
-    __slots__ = ["matches", "original_read"]
+    __slots__ = ["matches", "original_read", "cut_prefix", "cut_suffix", "is_rc"]
 
     def __init__(self, read):
         self.matches = []  # type: List[Match]
         self.original_read = read
+        self.cut_prefix = None
+        self.cut_suffix = None
+        self.is_rc = None
 
 
 class SingleEndModifier(ABC):
@@ -32,13 +44,15 @@ class SingleEndModifier(ABC):
         pass
 
 
-class PairedModifier(ABC):
+class PairedEndModifier(ABC):
     @abstractmethod
-    def __call__(self, read1, read2, info1: ModificationInfo, info2: ModificationInfo):
+    def __call__(
+        self, read1, read2, info1: ModificationInfo, info2: ModificationInfo
+    ) -> Tuple[DnaSequence, DnaSequence]:
         pass
 
 
-class PairedModifierWrapper(PairedModifier):
+class PairedEndModifierWrapper(PairedEndModifier):
     """
     Wrapper for modifiers that work on both reads in a paired-end read
     """
@@ -52,7 +66,7 @@ class PairedModifierWrapper(PairedModifier):
             raise ValueError("Not both modifiers may be None")
 
     def __repr__(self):
-        return 'PairedModifier(modifier1={!r}, modifier2={!r})'.format(
+        return "PairedEndModifierWrapper(modifier1={!r}, modifier2={!r})".format(
             self._modifier1, self._modifier2)
 
     def __call__(self, read1, read2, info1: ModificationInfo, info2: ModificationInfo):
@@ -70,38 +84,61 @@ class AdapterCutter(SingleEndModifier):
     times parameter.
     """
 
-    def __init__(self, adapters, times=1, action='trim'):
+    def __init__(
+        self,
+        adapters: List[Adapter],
+        times: int = 1,
+        action: Optional[str] = "trim",
+        index: bool = True,
+    ):
         """
-        adapters -- list of Adapter objects
+        action -- What to do with a found adapter:
+          None: Do nothing, only update the ModificationInfo appropriately
+          "trim": Remove the adapter and down- or upstream sequence depending on adapter type
+          "mask": Replace the part of the sequence that would have been removed with "N" bases
+          "lowercase": Convert the part of the sequence that would have been removed to lowercase
+          "retain": Like "trim", but leave the adapter sequence itself in the read
 
-        action -- What to do with a found adapter: None, 'trim', or 'mask'
+        index -- if True, an adapter index (for multiple adapters) is created if possible
         """
         self.times = times
-        assert action in ('trim', 'mask', 'lowercase', None)
+        assert action in ("trim", "mask", "lowercase", "retain", None)
         self.action = action
         self.with_adapters = 0
         self.adapter_statistics = OrderedDict((a, a.create_statistics()) for a in adapters)
-        prefix, suffix, other = self._split_adapters(adapters)
-        # For somewhat better backwards compatibility, avoid re-ordering
-        # the adapters when we don’t need to
-        if len(prefix) > 1 or len(suffix) > 1:
-            adapters = other
-            if len(prefix) > 1:
-                adapters.append(MultiPrefixAdapter(prefix))
-            else:
-                adapters.extend(prefix)
-            if len(suffix) > 1:
-                adapters.append(MultiSuffixAdapter(suffix))
-            else:
-                adapters.extend(suffix)
-        self.adapters = adapters
+        if index:
+            self.adapters = MultipleAdapters(self._regroup_into_indexed_adapters(adapters))
+        else:
+            self.adapters = MultipleAdapters(adapters)
+        if action == "retain" and times > 1:
+            raise ValueError("'retain' cannot be combined with times > 1")
 
     def __repr__(self):
         return 'AdapterCutter(adapters={!r}, times={}, action={!r})'.format(
             self.adapters, self.times, self.action)
 
+    def _regroup_into_indexed_adapters(self, adapters):
+        prefix, suffix, single = self._split_adapters(adapters)
+        # For somewhat better backwards compatibility, avoid re-ordering
+        # the adapters when we don’t need to
+        if len(prefix) > INDEXING_THRESHOLD or len(suffix) > INDEXING_THRESHOLD:
+            result = single
+            if len(prefix) > 1:
+                result.append(IndexedPrefixAdapters(prefix))
+            else:
+                result.extend(prefix)
+            if len(suffix) > 1:
+                result.append(IndexedSuffixAdapters(suffix))
+            else:
+                result.extend(suffix)
+            return result
+        else:
+            return adapters
+
     @staticmethod
-    def _split_adapters(adapters):
+    def _split_adapters(
+        adapters: Sequence[SingleAdapter]
+    ) -> Tuple[Sequence[SingleAdapter], Sequence[SingleAdapter], Sequence[SingleAdapter]]:
         """
         Split adapters into three different categories so that they can possibly be used
         with a MultiAdapter. Return a tuple (prefix, suffix, other), where
@@ -109,36 +146,22 @@ class AdapterCutter(SingleEndModifier):
         - suffix is a list of all anchored 3' adapters that MultiAdapter would accept
         - other is a list of all remaining adapters.
         """
-        prefix, suffix, other = [], [], []
+        prefix = []  # type: List[SingleAdapter]
+        suffix = []  # type: List[SingleAdapter]
+        other = []  # type: List[SingleAdapter]
         for a in adapters:
-            if MultiPrefixAdapter.is_acceptable(a):
-                lst = prefix
-            elif MultiSuffixAdapter.is_acceptable(a):
-                lst = suffix
+            if IndexedPrefixAdapters.is_acceptable(a):
+                prefix.append(a)
+            elif IndexedSuffixAdapters.is_acceptable(a):
+                suffix.append(a)
             else:
-                lst = other
-            lst.append(a)
+                other.append(a)
         return prefix, suffix, other
 
     @staticmethod
-    def best_match(adapters, read):
-        """
-        Find the best matching adapter in the given read.
-
-        Return either a Match instance or None if there are no matches.
-        """
-        best_match = None
-        for adapter in adapters:
-            match = adapter.match_to(read.sequence)
-            if match is None:
-                continue
-
-            # the no. of matches determines which adapter fits best
-            if best_match is None or match.matches > best_match.matches or (
-                match.matches == best_match.matches and match.errors < best_match.errors
-            ):
-                best_match = match
-        return best_match
+    def trim_but_retain_adapter(read, matches: Sequence[Match]):
+        start, stop = matches[-1].retained_adapter_interval()
+        return read[start:stop]
 
     @staticmethod
     def masked_read(read, matches: Sequence[Match]):
@@ -173,7 +196,7 @@ class AdapterCutter(SingleEndModifier):
     def match_and_trim(self, read):
         """
         Search for the best-matching adapter in a read, perform the requested action
-        ('trim', 'mask', 'lowercase' or None as determined by self.action) and return the
+        ('trim', 'mask' etc. as determined by self.action) and return the
         (possibly) modified read.
 
         *self.times* adapter removal rounds are done. During each round,
@@ -183,11 +206,11 @@ class AdapterCutter(SingleEndModifier):
         Return a pair (trimmed_read, matches), where matches is a list of Match instances.
         """
         matches = []
-        if self.action == 'lowercase':
+        if self.action == 'lowercase':  # TODO this should not be needed
             read.sequence = read.sequence.upper()
         trimmed_read = read
         for _ in range(self.times):
-            match = AdapterCutter.best_match(self.adapters, trimmed_read)
+            match = self.adapters.match_to(trimmed_read.sequence)
             if match is None:
                 # if nothing found, attempt no further rounds
                 break
@@ -200,12 +223,14 @@ class AdapterCutter(SingleEndModifier):
         if self.action == 'trim':
             # read is already trimmed, nothing to do
             pass
+        elif self.action == 'retain':
+            trimmed_read = self.trim_but_retain_adapter(read, matches)
         elif self.action == 'mask':
             trimmed_read = self.masked_read(read, matches)
         elif self.action == 'lowercase':
             trimmed_read = self.lowercased_read(read, matches)
             assert len(trimmed_read.sequence) == len(read)
-        elif self.action is None:  # --no-trim
+        elif self.action is None:
             trimmed_read = read[:]
 
         return trimmed_read, matches
@@ -236,9 +261,11 @@ class ReverseComplementer(SingleEndModifier):
             self.reverse_complemented += 1
             assert reverse_matches
             trimmed_read, matches = reverse_trimmed_read, reverse_matches
+            info.is_rc = True
             if self._suffix:
                 trimmed_read.name += self._suffix
         else:
+            info.is_rc = False
             trimmed_read, matches = forward_trimmed_read, forward_matches
 
         if matches:
@@ -255,7 +282,7 @@ class PairedAdapterCutterError(Exception):
     pass
 
 
-class PairedAdapterCutter(PairedModifier):
+class PairedAdapterCutter(PairedEndModifier):
     """
     A Modifier that trims adapter pairs from R1 and R2.
     """
@@ -277,9 +304,9 @@ class PairedAdapterCutter(PairedModifier):
                 "Given: {} for R1, {} for R2".format(len(adapters1), len(adapters2)))
         if not adapters1:
             raise PairedAdapterCutterError("No adapters given")
-        self._adapters1 = adapters1
+        self._adapters1 = MultipleAdapters(adapters1)
         self._adapter_indices = {a: i for i, a in enumerate(adapters1)}
-        self._adapters2 = adapters2
+        self._adapters2 = MultipleAdapters(adapters2)
         self.action = action
         self.with_adapters = 0
         self.adapter_statistics = [None, None]
@@ -293,7 +320,7 @@ class PairedAdapterCutter(PairedModifier):
     def __call__(self, read1, read2, info1, info2):
         """
         """
-        match1 = AdapterCutter.best_match(self._adapters1, read1)
+        match1 = self._adapters1.match_to(read1.sequence)
         if match1 is None:
             return read1, read2
         adapter1 = match1.adapter
@@ -320,6 +347,8 @@ class PairedAdapterCutter(PairedModifier):
             elif self.action == 'lowercase':
                 trimmed_read = AdapterCutter.lowercased_read(read, [match])
                 assert len(trimmed_read.sequence) == len(read)
+            elif self.action == 'retain':
+                trimmed_read = AdapterCutter.trim_but_retain_adapter(read, [match])
             elif self.action is None:  # --no-trim
                 trimmed_read = read[:]
             result.append(trimmed_read)
@@ -340,8 +369,10 @@ class UnconditionalCutter(SingleEndModifier):
 
     def __call__(self, read, info: ModificationInfo):
         if self.length > 0:
+            info.cut_prefix = read.sequence[:self.length]
             return read[self.length:]
         elif self.length < 0:
+            info.cut_suffix = read.sequence[self.length:]
             return read[:self.length]
 
 
@@ -390,6 +421,174 @@ class PrefixSuffixAdder(SingleEndModifier):
         return read
 
 
+class InvalidTemplate(Exception):
+    pass
+
+
+class Renamer(SingleEndModifier):
+    """
+    Rename reads using a template
+
+    The template string can contain the following placeholders:
+
+    - {header} -- full, unchanged header
+    - {id} -- the part of the header before the first whitespace
+    - {comment} -- the part of the header after the ID, excluding initial whitespace
+    - {cut_prefix} -- prefix removed by UnconditionalCutter (with positive length argument)
+    - {cut_suffix} -- suffix removed by UnconditionalCutter (with negative length argument)
+    - {adapter_name} -- name of the *last* adapter match or no_adapter if there was none
+    """
+    variables = {
+        "header",
+        "id",
+        "comment",
+        "cut_prefix",
+        "cut_suffix",
+        "adapter_name",
+    }
+
+    def __init__(self, template: str):
+        try:
+            self._tokens = list(tokenize_braces(template))
+        except TokenizeError as e:
+            raise InvalidTemplate("Error in template '{}': {}".format(template, e))
+        self.raise_if_invalid_variable(self._tokens, self.variables)
+        self._template = template
+
+    @staticmethod
+    def raise_if_invalid_variable(tokens: List[Token], allowed: Set[str]) -> None:
+        for token in tokens:
+            if not isinstance(token, BraceToken):
+                continue
+            value = token.value
+            if value not in allowed:
+                raise InvalidTemplate(
+                    "Error in template: Variable '{}' not recognized".format(value)
+                )
+
+    @staticmethod
+    def parse_name(read_name: str) -> Tuple[str, str]:
+        """Parse read header and return (id, comment) tuple"""
+        fields = read_name.split(maxsplit=1)
+        if len(fields) == 2:
+            return (fields[0], fields[1])
+        else:
+            return (read_name, "")
+
+    def __call__(self, read: DnaSequence, info: ModificationInfo) -> DnaSequence:
+        id_, comment = self.parse_name(read.name)
+        read.name = self._template.format(
+            header=read.name,
+            id=id_,
+            comment=comment,
+            cut_prefix=info.cut_prefix if info.cut_prefix else "",
+            cut_suffix=info.cut_suffix if info.cut_suffix else "",
+            adapter_name=info.matches[-1].adapter.name if info.matches else "no_adapter",
+        )
+        return read
+
+
+class PairedEndRenamer(PairedEndModifier):
+    """
+    Rename paired-end reads using a template. The template is applied to both
+    R1 and R2, and the same template variables as in the (single-end) renamer
+    are allowed. However,
+    these variables are evaluated separately for each read. For example, if `{comment}`
+    is used, it gets replaced with the R1 comment in the R1 header, and with the R2
+    comment in the R2 header.
+
+    Additionally, all template variables except `id` can be used in the read-specific
+    forms `{r1.variablename}` and `{r2.variablename}`. For example, `{r1.comment}`
+    always gets replaced with the R1 comment, even in R2.
+    """
+
+    def __init__(self, template: str):
+        try:
+            self._tokens = list(tokenize_braces(template))
+        except TokenizeError as e:
+            raise InvalidTemplate("Error in template '{}': {}".format(template, e))
+        Renamer.raise_if_invalid_variable(self._tokens, self._get_allowed_variables())
+        self._template = template
+
+    @staticmethod
+    def _get_allowed_variables() -> Set[str]:
+        allowed = Renamer.variables.copy()
+        allowed.add("rn")
+        for v in Renamer.variables - {"id"}:
+            allowed.add("r1." + v)
+            allowed.add("r2." + v)
+        return allowed
+
+    def __call__(
+        self, read1: DnaSequence, read2: DnaSequence, info1: ModificationInfo, info2: ModificationInfo
+    ) -> Tuple[DnaSequence, DnaSequence]:
+
+        id1, comment1 = Renamer.parse_name(read1.name)
+        id2, comment2 = Renamer.parse_name(read2.name)
+        if not record_names_match(read1.name, read2.name):
+            raise ValueError("Input read IDs not identical: '{}' != '{}'".format(id1, id2))
+        name1, name2 = self.get_new_headers(
+            id1=id1,
+            id2=id2,
+            comment1=comment1,
+            comment2=comment2,
+            header1=read1.name,
+            header2=read2.name,
+            info1=info1,
+            info2=info2,
+        )
+        new_id1 = Renamer.parse_name(name1)[0]
+        new_id2 = Renamer.parse_name(name2)[0]
+        if not record_names_match(name1, name2):
+            raise InvalidTemplate(
+                "After renaming R1 and R2, their IDs are no longer identical: "
+                "'{}' != '{}'. Original read ID: '{}'. ".format(new_id1, new_id2, id1)
+            )
+        read1.name = name1
+        read2.name = name2
+        return read1, read2
+
+    def get_new_headers(
+        self,
+        id1: str,
+        id2: str,
+        comment1: str,
+        comment2: str,
+        header1: str,
+        header2: str,
+        info1: ModificationInfo,
+        info2: ModificationInfo,
+    ) -> Tuple[str, str]:
+        d = []
+        for id_, comment, header, info in (
+            (id1, comment1, header1, info1), (id2, comment2, header2, info2)
+        ):
+            d.append(
+                dict(
+                    comment=comment,
+                    header=header,
+                    cut_prefix=info.cut_prefix if info.cut_prefix else "",
+                    cut_suffix=info.cut_suffix if info.cut_suffix else "",
+                    adapter_name=info.matches[-1].adapter.name if info.matches else "no_adapter",
+                )
+            )
+        name1 = self._template.format(
+            id=id1,
+            rn=1,
+            **d[0],
+            r1=SimpleNamespace(**d[0]),
+            r2=SimpleNamespace(**d[1]),
+        )
+        name2 = self._template.format(
+            id=id2,
+            rn=2,
+            **d[1],
+            r1=SimpleNamespace(**d[0]),
+            r2=SimpleNamespace(**d[1]),
+        )
+        return name1, name2
+
+
 class ZeroCapper(SingleEndModifier):
     """
     Change negative quality values of a read to zero
