@@ -1,25 +1,29 @@
 """
 Routines for printing a report.
 """
+import sys
 from io import StringIO
 import textwrap
-from typing import Any, Optional, List
-from .adapters import Where, EndStatistics, AdapterStatistics, ADAPTER_TYPE_NAMES
-from .modifiers import (Modifier, PairedModifier, QualityTrimmer, NextseqQualityTrimmer,
-    AdapterCutter, PairedAdapterCutter, ReverseComplementer)
-from .filters import (NoFilter, PairedNoFilter, TooShortReadFilter, TooLongReadFilter,
-    PairedDemultiplexer, CombinatorialDemultiplexer, Demultiplexer, NContentFilter, InfoFileWriter,
-    WildcardFileWriter, RestFileWriter)
+from collections import Counter
+from typing import Any, Optional, List, Dict
+from .adapters import (
+    EndStatistics, AdapterStatistics, FrontAdapter, NonInternalFrontAdapter, PrefixAdapter,
+    BackAdapter, NonInternalBackAdapter, SuffixAdapter, AnywhereAdapter, LinkedAdapter,
+)
+from .modifiers import (QualityTrimmer, NextseqQualityTrimmer,
+    AdapterCutter, PairedAdapterCutter, ReverseComplementer, PairedEndModifierWrapper)
+from .filters import (WithStatistics, TooShortReadFilter, TooLongReadFilter, NContentFilter,
+    CasavaFilter, MaximumExpectedErrorsFilter)
 
 
-def safe_divide(numerator, denominator):
+def safe_divide(numerator: Optional[int], denominator: int) -> float:
     if numerator is None or not denominator:
         return 0.0
     else:
         return numerator / denominator
 
 
-def add_if_not_none(a, b):
+def add_if_not_none(a: Optional[int], b: Optional[int]) -> Optional[int]:
     if a is None:
         return b
     if b is None:
@@ -31,19 +35,22 @@ class Statistics:
     def __init__(self) -> None:
         """
         """
-        self.paired = None  # type: Optional[bool]
-        self.did_quality_trimming = None  # type: Optional[bool]
-        self.too_short = None
-        self.too_long = None
-        self.too_many_n = None
-        self.reverse_complemented = None  # type: Optional[int]
+        self.paired: Optional[bool] = None
+        self.did_quality_trimming: Optional[bool] = None
+        self.too_short: Optional[int] = None
+        self.too_long: Optional[int] = None
+        self.too_many_n: Optional[int] = None
+        self.too_many_expected_errors: Optional[int] = None
+        self.casava_filtered: Optional[int] = None
+        self.reverse_complemented: Optional[int] = None
         self.n = 0
         self.written = 0
         self.total_bp = [0, 0]
         self.written_bp = [0, 0]
+        self.written_lengths: List[Counter] = [Counter(), Counter()]
         self.with_adapters = [0, 0]
         self.quality_trimmed_bp = [0, 0]
-        self.adapter_stats = [[], []]  # type: List[List[AdapterStatistics]]
+        self.adapter_stats: List[List[AdapterStatistics]] = [[], []]
 
     def __iadd__(self, other: Any):
         self.n += other.n
@@ -63,9 +70,13 @@ class Statistics:
         self.too_short = add_if_not_none(self.too_short, other.too_short)
         self.too_long = add_if_not_none(self.too_long, other.too_long)
         self.too_many_n = add_if_not_none(self.too_many_n, other.too_many_n)
+        self.too_many_expected_errors = add_if_not_none(
+            self.too_many_expected_errors, other.too_many_expected_errors)
+        self.casava_filtered = add_if_not_none(self.casava_filtered, other.casava_filtered)
         for i in (0, 1):
             self.total_bp[i] += other.total_bp[i]
             self.written_bp[i] += other.written_bp[i]
+            self.written_lengths[i] += other.written_lengths[i]
             self.with_adapters[i] += other.with_adapters[i]
             self.quality_trimmed_bp[i] += other.quality_trimmed_bp[i]
             if self.adapter_stats[i] and other.adapter_stats[i]:
@@ -78,7 +89,7 @@ class Statistics:
                 self.adapter_stats[i] = other.adapter_stats[i]
         return self
 
-    def collect(self, n, total_bp1, total_bp2, modifiers, writers):
+    def collect(self, n: int, total_bp1: int, total_bp2: Optional[int], modifiers, writers):
         """
         n -- total number of reads
         total_bp1 -- number of bases in first reads
@@ -101,28 +112,33 @@ class Statistics:
         # For chaining
         return self
 
-    def _collect_writer(self, w):
-        if isinstance(w, (InfoFileWriter, RestFileWriter, WildcardFileWriter)):
-            return
-        elif isinstance(w, (NoFilter, PairedNoFilter, PairedDemultiplexer,
-                CombinatorialDemultiplexer, Demultiplexer)):
-            self.written += w.written
-            self.written_bp[0] += w.written_bp[0]
-            self.written_bp[1] += w.written_bp[1]
-        elif isinstance(w.filter, TooShortReadFilter):
-            self.too_short = w.filtered
-        elif isinstance(w.filter, TooLongReadFilter):
-            self.too_long = w.filtered
-        elif isinstance(w.filter, NContentFilter):
-            self.too_many_n = w.filtered
+    def _collect_writer(self, w) -> None:
+        if isinstance(w, WithStatistics):
+            self.written += w.written_reads()
+            written_bp = w.written_bp()
+            written_lengths = w.written_lengths()
+            for i in 0, 1:
+                self.written_bp[i] += written_bp[i]
+                self.written_lengths[i] += written_lengths[i]
+        if hasattr(w, "filter"):
+            if isinstance(w.filter, TooShortReadFilter):
+                self.too_short = w.filtered
+            elif isinstance(w.filter, TooLongReadFilter):
+                self.too_long = w.filtered
+            elif isinstance(w.filter, NContentFilter):
+                self.too_many_n = w.filtered
+            elif isinstance(w.filter, MaximumExpectedErrorsFilter):
+                self.too_many_expected_errors = w.filtered
+            elif isinstance(w.filter, CasavaFilter):
+                self.casava_filtered = w.filtered
 
-    def _collect_modifier(self, m: Modifier):
+    def _collect_modifier(self, m) -> None:
         if isinstance(m, PairedAdapterCutter):
             for i in 0, 1:
                 self.with_adapters[i] += m.with_adapters
                 self.adapter_stats[i] = list(m.adapter_statistics[i].values())
             return
-        if isinstance(m, PairedModifier):
+        if isinstance(m, PairedEndModifierWrapper):
             modifiers_list = [(0, m._modifier1), (1, m._modifier2)]
         else:
             modifiers_list = [(0, m)]
@@ -139,54 +155,62 @@ class Statistics:
                 self.reverse_complemented = modifier.reverse_complemented
 
     @property
-    def total(self):
+    def total(self) -> int:
         return sum(self.total_bp)
 
     @property
-    def quality_trimmed(self):
+    def quality_trimmed(self) -> int:
         return sum(self.quality_trimmed_bp)
 
     @property
-    def total_written_bp(self):
+    def total_written_bp(self) -> int:
         return sum(self.written_bp)
 
     @property
-    def written_fraction(self):
+    def written_fraction(self) -> float:
         return safe_divide(self.written, self.n)
 
     @property
-    def with_adapters_fraction(self):
+    def with_adapters_fraction(self) -> List[float]:
         return [safe_divide(v, self.n) for v in self.with_adapters]
 
     @property
-    def quality_trimmed_fraction(self):
+    def quality_trimmed_fraction(self) -> float:
         return safe_divide(self.quality_trimmed, self.total)
 
     @property
-    def total_written_bp_fraction(self):
+    def total_written_bp_fraction(self) -> float:
         return safe_divide(self.total_written_bp, self.total)
 
     @property
-    def reverse_complemented_fraction(self):
+    def reverse_complemented_fraction(self) -> float:
         return safe_divide(self.reverse_complemented, self.n)
 
     @property
-    def too_short_fraction(self):
+    def too_short_fraction(self) -> float:
         return safe_divide(self.too_short, self.n)
 
     @property
-    def too_long_fraction(self):
+    def too_long_fraction(self) -> float:
         return safe_divide(self.too_long, self.n)
 
     @property
-    def too_many_n_fraction(self):
+    def too_many_n_fraction(self) -> float:
         return safe_divide(self.too_many_n, self.n)
 
+    @property
+    def too_many_expected_errors_fraction(self) -> float:
+        return safe_divide(self.too_many_expected_errors, self.n)
 
-def error_ranges(adapter_statistics: EndStatistics):
+    @property
+    def casava_filtered_fraction(self) -> float:
+        return safe_divide(self.casava_filtered, self.n)
+
+
+def error_ranges(adapter_statistics: EndStatistics) -> str:
     length = adapter_statistics.effective_length
     error_rate = adapter_statistics.max_error_rate
-    prev = 0
+    prev = 1
     s = ""
     for errors in range(1, int(error_rate * length) + 1):
         r = int(errors / error_rate)
@@ -200,7 +224,7 @@ def error_ranges(adapter_statistics: EndStatistics):
     return "No. of allowed errors:\n" + s + "\n"
 
 
-def histogram(end_statistics: EndStatistics, n: int, gc_content: float):
+def histogram(end_statistics: EndStatistics, n: int, gc_content: float) -> str:
     """
     Return a formatted histogram. Include the no. of reads expected to be
     trimmed by chance (assuming a uniform distribution of nucleotides in the reads).
@@ -235,11 +259,11 @@ def histogram(end_statistics: EndStatistics, n: int, gc_content: float):
 
 
 class AdjacentBaseStatistics:
-    def __init__(self, bases):
+    def __init__(self, bases: Dict[str, int]):
         """
         """
-        self.bases = bases
-        self._warnbase = None
+        self.bases: Dict[str, int] = bases
+        self._warnbase: Optional[str] = None
         total = sum(self.bases.values())
         if total == 0:
             self._fractions = None
@@ -258,10 +282,10 @@ class AdjacentBaseStatistics:
         return 'AdjacentBaseStatistics(bases={})'.format(self.bases)
 
     @property
-    def should_warn(self):
+    def should_warn(self) -> bool:
         return self._warnbase is not None
 
-    def __str__(self):
+    def __str__(self) -> str:
         if not self._fractions:
             return ""
         sio = StringIO()
@@ -271,7 +295,7 @@ class AdjacentBaseStatistics:
         if self.should_warn:
             print('WARNING:', file=sio)
             print('    The adapter is preceded by "{}" extremely often.'.format(self._warnbase), file=sio)
-            print("    The provided adapter sequence could be incomplete at its 3' end.", file=sio)
+            print("    The provided adapter sequence could be incomplete at its 5' end.", file=sio)
         return sio.getvalue()
 
 
@@ -286,8 +310,12 @@ def full_report(stats: Statistics, time: float, gc_content: float) -> str:  # no
         kwargs['file'] = sio
         print(*args, **kwargs)
 
-    print_s("Finished in {:.2F} s ({:.0F} us/read; {:.2F} M reads/minute).".format(
-        time, 1E6 * time / stats.n, stats.n / time * 60 / 1E6))
+    if sys.version_info[:2] <= (3, 6):
+        micro = "u"
+    else:
+        micro = "µ"
+    print_s("Finished in {:.2F} s ({:.0F} {}s/read; {:.2F} M reads/minute).".format(
+        time, 1E6 * time / stats.n, micro, stats.n / time * 60 / 1E6))
 
     report = "\n=== Summary ===\n\n"
     if stats.paired:
@@ -304,12 +332,19 @@ def full_report(stats: Statistics, time: float, gc_content: float) -> str:  # no
     if stats.reverse_complemented is not None:
         report += "Reverse-complemented:            " \
                   "{o.reverse_complemented:13,d} ({o.reverse_complemented_fraction:.1%})\n"
+
     if stats.too_short is not None:
         report += "{pairs_or_reads} that were too short:       {o.too_short:13,d} ({o.too_short_fraction:.1%})\n"
     if stats.too_long is not None:
         report += "{pairs_or_reads} that were too long:        {o.too_long:13,d} ({o.too_long_fraction:.1%})\n"
     if stats.too_many_n is not None:
         report += "{pairs_or_reads} with too many N:           {o.too_many_n:13,d} ({o.too_many_n_fraction:.1%})\n"
+    if stats.too_many_expected_errors is not None:
+        report += "{pairs_or_reads} with too many exp. errors: " \
+                  "{o.too_many_expected_errors:13,d} ({o.too_many_expected_errors_fraction:.1%})\n"
+    if stats.casava_filtered is not None:
+        report += "{pairs_or_reads} failed CASAVA filter:      " \
+                  "{o.casava_filtered:13,d} ({o.casava_filtered_fraction:.1%})\n"
 
     report += textwrap.dedent("""\
     {pairs_or_reads} written (passing filters): {o.written:13,d} ({o.written_fraction:.1%})
@@ -341,12 +376,11 @@ def full_report(stats: Statistics, time: float, gc_content: float) -> str:  # no
             total_back = sum(adapter_statistics.back.lengths.values())
             total = total_front + total_back
             reverse_complemented = adapter_statistics.reverse_complemented
-            where = adapter_statistics.where
-            where_backs = (Where.BACK, Where.BACK_NOT_INTERNAL, Where.SUFFIX)
-            where_fronts = (Where.FRONT, Where.FRONT_NOT_INTERNAL, Where.PREFIX)
-            assert (where in (Where.ANYWHERE, Where.LINKED)
-                or (where in where_backs and total_front == 0)
-                or (where in where_fronts and total_back == 0)), (where, total_front, total_back)
+            adapter = adapter_statistics.adapter
+            if isinstance(adapter, (BackAdapter, NonInternalBackAdapter, SuffixAdapter)):
+                assert total_front == 0
+            if isinstance(adapter, (FrontAdapter, NonInternalFrontAdapter, PrefixAdapter)):
+                assert total_back == 0
 
             if stats.paired:
                 extra = 'First read: ' if which_in_pair == 0 else 'Second read: '
@@ -356,7 +390,7 @@ def full_report(stats: Statistics, time: float, gc_content: float) -> str:  # no
             print_s("=" * 3, extra + "Adapter", adapter_statistics.name, "=" * 3)
             print_s()
 
-            if where is Where.LINKED:
+            if isinstance(adapter, LinkedAdapter):
                 print_s("Sequence: {}...{}; Type: linked; Length: {}+{}; "
                     "5' trimmed: {} times; 3' trimmed: {} times".format(
                         adapter_statistics.front.sequence,
@@ -366,16 +400,16 @@ def full_report(stats: Statistics, time: float, gc_content: float) -> str:  # no
                         total_front, total_back))
             else:
                 print_s("Sequence: {}; Type: {}; Length: {}; Trimmed: {} times".
-                    format(adapter_statistics.front.sequence, ADAPTER_TYPE_NAMES[adapter_statistics.where],
+                    format(adapter_statistics.front.sequence, adapter.description,
                         len(adapter_statistics.front.sequence), total), end="")
-            if reverse_complemented is not None:
+            if stats.reverse_complemented is not None:
                 print_s("; Reverse-complemented: {} times".format(reverse_complemented))
             else:
                 print_s()
             if total == 0:
                 print_s()
                 continue
-            if where is Where.ANYWHERE:
+            if isinstance(adapter, AnywhereAdapter):
                 print_s(total_front, "times, it overlapped the 5' end of a read")
                 print_s(total_back, "times, it overlapped the 3' end or was within the read")
                 print_s()
@@ -385,7 +419,7 @@ def full_report(stats: Statistics, time: float, gc_content: float) -> str:  # no
                 print_s()
                 print_s("Overview of removed sequences (3' or within)")
                 print_s(histogram(adapter_statistics.back, stats.n, gc_content))
-            elif where is Where.LINKED:
+            elif isinstance(adapter, LinkedAdapter):
                 print_s()
                 print_s(error_ranges(adapter_statistics.front))
                 print_s(error_ranges(adapter_statistics.back))
@@ -394,13 +428,13 @@ def full_report(stats: Statistics, time: float, gc_content: float) -> str:  # no
                 print_s()
                 print_s("Overview of removed sequences at 3' end")
                 print_s(histogram(adapter_statistics.back, stats.n, gc_content))
-            elif where in (Where.FRONT, Where.PREFIX, Where.FRONT_NOT_INTERNAL):
+            elif isinstance(adapter, (FrontAdapter, NonInternalFrontAdapter, PrefixAdapter)):
                 print_s()
                 print_s(error_ranges(adapter_statistics.front))
                 print_s("Overview of removed sequences")
                 print_s(histogram(adapter_statistics.front, stats.n, gc_content))
             else:
-                assert where in (Where.BACK, Where.SUFFIX, Where.BACK_NOT_INTERNAL)
+                assert isinstance(adapter, (BackAdapter, NonInternalBackAdapter, SuffixAdapter))
                 print_s()
                 print_s(error_ranges(adapter_statistics.back))
                 base_stats = AdjacentBaseStatistics(adapter_statistics.back.adjacent_bases)
@@ -417,8 +451,10 @@ def full_report(stats: Statistics, time: float, gc_content: float) -> str:  # no
     return sio.getvalue().rstrip()
 
 
-def minimal_report(stats, _time, _gc_content) -> str:
+def minimal_report(stats: Statistics, time: float, gc_content: float) -> str:
     """Create a minimal tabular report suitable for concatenation"""
+    _ = time
+    _ = gc_content
 
     def none(value):
         return 0 if value is None else value
@@ -445,7 +481,7 @@ def minimal_report(stats, _time, _gc_content) -> str:
     warning = False
     for which_in_pair in (0, 1):
         for adapter_statistics in stats.adapter_stats[which_in_pair]:
-            if adapter_statistics.where in (Where.BACK, Where.SUFFIX, Where.BACK_NOT_INTERNAL):
+            if isinstance(adapter_statistics.adapter, (BackAdapter, NonInternalBackAdapter, SuffixAdapter)):
                 if AdjacentBaseStatistics(adapter_statistics.back.adjacent_bases).should_warn:
                     warning = True
                     break
