@@ -1,12 +1,7 @@
-# TODO
-# test reading from standard input
-import os
-import shutil
-import sys
-import tempfile
-from io import StringIO
-import pytest
 import subprocess
+import sys
+from io import StringIO, BytesIO
+import pytest
 
 from cutadapt.__main__ import main
 from utils import assert_files_equal, datapath, cutpath
@@ -19,6 +14,46 @@ except ImportError:  # pragma: no cover
 del _unused
 
 
+def test_does_not_close_stdout():
+    main([datapath("small.fastq")])
+    assert not sys.stdout.closed
+
+
+def test_help():
+    with pytest.raises(SystemExit) as e:
+        main(["--help"])
+    assert e.value.args[0] == 0
+
+
+def test_unknown_file_format(tmp_path):
+    path = tmp_path / "unknown_format.txt"
+    path.write_text("raw text")
+    with pytest.raises(SystemExit):
+        main([str(path)])
+
+
+def test_cores_negative():
+    with pytest.raises(SystemExit) as e:
+        main(["--cores=-1", datapath("simple.fasta")])
+    assert e.value.args[0] == 2
+    # "cannot be negative"
+
+
+def test_quiet_and_report():
+    with pytest.raises(SystemExit) as e:
+        main(["--quiet", "--report=minimal", datapath("simple.fasta")])
+    assert e.value.args[0] == 2
+    # "Options --quiet and --report cannot be used at the same time"
+
+
+def test_debug():
+    main(["--debug", "--", datapath("small.fastq")])
+
+
+def test_debug_trace():
+    main(["--debug", "--debug", "-a", "ACGT", datapath("small.fastq")])
+
+
 def test_example(run):
     run('-N -b ADAPTER', 'example.fa', 'example.fa')
 
@@ -74,11 +109,6 @@ def test_extensiontxtgz(run):
     run("-b TTAGACATATCTCCGTCG", "s_1_sequence.txt", "s_1_sequence.txt.gz")
 
 
-def test_format(run):
-    """the -f/--format parameter"""
-    run("-f fastq -b TTAGACATATCTCCGTCG", "small.fastq", "small.myownextension")
-
-
 def test_minimum_length(run):
     """-m/--minimum-length"""
     run("-m 5 -a TTAGACATATCTCCGTCG", "minlen.fa", "lengths.fa")
@@ -87,13 +117,14 @@ def test_minimum_length(run):
 def test_too_short(run, tmpdir, cores):
     """--too-short-output"""
     too_short_path = str(tmpdir.join('tooshort.fa'))
-    run([
+    stats = run([
         "--cores", str(cores),
         "-m", "5",
         "-a", "TTAGACATATCTCCGTCG",
         "--too-short-output", too_short_path
     ], "minlen.fa", "lengths.fa")
     assert_files_equal(datapath('tooshort.fa'), too_short_path)
+    assert stats.too_short == 5
 
 
 def test_maximum_length(run):
@@ -104,13 +135,14 @@ def test_maximum_length(run):
 def test_too_long(run, tmpdir, cores):
     """--too-long-output"""
     too_long_path = str(tmpdir.join('toolong.fa'))
-    run([
+    stats = run([
         "--cores", str(cores),
         "-M", "5",
         "-a", "TTAGACATATCTCCGTCG",
         "--too-long-output", too_long_path
     ], "maxlen.fa", "lengths.fa")
     assert_files_equal(datapath('toolong.fa'), too_long_path)
+    assert stats.too_long == 5
 
 
 def test_length_tag(run):
@@ -195,17 +227,20 @@ def test_action_lowercase(run):
     run("-b CAAG -n 3 --action=lowercase", "action_lowercase.fasta", "action_lowercase.fasta")
 
 
+def test_action_retain(run):
+    run("-g GGTTAACC -a CAAG --action=retain", "action_retain.fasta", "action_retain.fasta")
+
+
+def test_action_retain_times():
+    with pytest.raises(SystemExit):
+        main(["-a", "ACGT", "--times=2", "--action=retain", datapath("small.fastq")])
+
+
 def test_gz_multiblock(run):
     """compressed gz file with multiple blocks (created by concatenating two .gz files)"""
     run("-b TTAGACATATCTCCGTCG", "small.fastq", "multiblock.fastq.gz")
 
 
-@pytest.mark.parametrize("opt", ["-y", "--suffix"])
-def test_suffix(opt, run):
-    """-y/--suffix parameter"""
-    run([opt, ' {name}', '-e', '0', '-a', 'OnlyT=TTTTTTTT', '-a', 'OnlyG=GGGGGGGG'], "suffix.fastq", "suffix.fastq")
-
-
 def test_read_wildcard(run):
     """test wildcards in reads"""
     run("--match-read-wildcards -b ACGTACGT", "wildcard.fa", "wildcard.fa")
@@ -339,7 +374,7 @@ def test_compressed_output(tmp_path, cores, extension):
     out_path = str(tmp_path / ("small.fastq." + extension))
     params = [
         "--cores", str(cores), "-a", "TTAGACATATCTCCGTCG", "-o", out_path, datapath("small.fastq")]
-    assert main(params) is None
+    main(params)
 
 
 if sys.version_info[:2] >= (3, 3):
@@ -435,15 +470,24 @@ def test_adapter_file_empty_name(run):
     run('-N -a file:' + datapath('adapter-empty-name.fasta'), 'illumina.fastq', 'illumina.fastq.gz')
 
 
-def test_demultiplex():
-    tempdir = tempfile.mkdtemp(prefix='cutadapt-tests.')
-    multiout = os.path.join(tempdir, 'tmp-demulti.{name}.fasta')
-    params = ['-a', 'first=AATTTCAGGAATT', '-a', 'second=GTTCTCTAGTTCT', '-o', multiout, datapath('twoadapters.fasta')]
-    assert main(params) is None
-    assert_files_equal(cutpath('twoadapters.first.fasta'), multiout.format(name='first'))
-    assert_files_equal(cutpath('twoadapters.second.fasta'), multiout.format(name='second'))
-    assert_files_equal(cutpath('twoadapters.unknown.fasta'), multiout.format(name='unknown'))
-    shutil.rmtree(tempdir)
+@pytest.mark.parametrize("ext", ["", ".gz"])
+def test_demultiplex(cores, tmp_path, ext):
+    multiout = str(tmp_path / 'tmp-demulti.{name}.fasta') + ext
+    params = [
+        '--cores', str(cores),
+        '-a', 'first=AATTTCAGGAATT',
+        '-a', 'second=GTTCTCTAGTTCT',
+        '-o', multiout,
+        datapath('twoadapters.fasta'),
+    ]
+    main(params)
+    for name in ("first", "second", "unknown"):
+        actual = multiout.format(name=name)
+        if ext == ".gz":
+            subprocess.run(["gzip", "-d", actual], check=True)
+            actual = actual[:-3]
+        expected = cutpath("twoadapters.{name}.fasta".format(name=name))
+        assert_files_equal(expected, actual)
 
 
 def test_multiple_fake_anchored_adapters(run):
@@ -467,27 +511,31 @@ def test_multiple_suffix_adapters_noindels(run):
 
 
 def test_max_n(run):
-    run('--max-n 0', 'maxn0.fasta', 'maxn.fasta')
-    run('--max-n 1', 'maxn1.fasta', 'maxn.fasta')
-    run('--max-n 2', 'maxn2.fasta', 'maxn.fasta')
-    run('--max-n 0.2', 'maxn0.2.fasta', 'maxn.fasta')
-    run('--max-n 0.4', 'maxn0.4.fasta', 'maxn.fasta')
+    assert run('--max-n 0', 'maxn0.fasta', 'maxn.fasta').too_many_n == 4
+    assert run('--max-n 1', 'maxn1.fasta', 'maxn.fasta').too_many_n == 2
+    assert run('--max-n 2', 'maxn2.fasta', 'maxn.fasta').too_many_n == 1
+    assert run('--max-n 0.2', 'maxn0.2.fasta', 'maxn.fasta').too_many_n == 3
+    assert run('--max-n 0.4', 'maxn0.4.fasta', 'maxn.fasta').too_many_n == 2
 
 
 def test_quiet_is_quiet():
     captured_standard_output = StringIO()
     captured_standard_error = StringIO()
+    setattr(captured_standard_output, "buffer", BytesIO())
+    setattr(captured_standard_error, "buffer", BytesIO())
     old_stdout = sys.stdout
     old_stderr = sys.stderr
     try:
         sys.stdout = captured_standard_output
         sys.stderr = captured_standard_error
-        main(['-o', '/dev/null', '--quiet', '-a', 'XXXX', datapath('illumina.fastq.gz')])
+        main(['-o', '/dev/null', '--quiet', datapath('small.fastq')])
     finally:
         sys.stdout = old_stdout
         sys.stderr = old_stderr
     assert captured_standard_output.getvalue() == ''
     assert captured_standard_error.getvalue() == ''
+    assert getattr(captured_standard_output, "buffer").getvalue() == b''
+    assert getattr(captured_standard_output, "buffer").getvalue() == b''
 
 
 def test_x_brace_notation():
@@ -565,10 +613,6 @@ def test_negative_length(run):
     run('--length -5', 'shortened-negative.fastq', 'small.fastq')
 
 
-def test_run_cutadapt_process():
-    subprocess.check_call(['cutadapt', '--version'])
-
-
 @pytest.mark.timeout(0.5)
 def test_issue_296(tmpdir):
     # Hang when using both --no-trim and --info-file together
@@ -591,7 +635,8 @@ def test_adapterx(run):
 
 
 def test_discard_casava(run):
-    run('--discard-casava', 'casava.fastq', 'casava.fastq')
+    stats = run('--discard-casava', 'casava.fastq', 'casava.fastq')
+    assert stats.casava_filtered == 1
 
 
 def test_underscore(run):
@@ -618,72 +663,6 @@ def test_paired_separate(run):
     run("-a CAGTGGAGTA", "paired-separate.2.fastq", "paired.2.fastq")
 
 
-def test_run_as_module():
-    """Check that "python3 -m cutadapt ..." works"""
-    import subprocess
-    from cutadapt import __version__
-    py = subprocess.Popen([sys.executable, "-m", "cutadapt", "--version"], stdout=subprocess.PIPE)
-    assert py.communicate()[0].decode().strip() == __version__
-
-
-def test_standard_input_pipe(tmpdir, cores):
-    """Read FASTQ from standard input"""
-
-    import subprocess
-    out_path = str(tmpdir.join("out.fastq"))
-    in_path = datapath("small.fastq")
-    # Use 'cat' to simulate that no file name is available for stdin
-    cat = subprocess.Popen(["cat", in_path], stdout=subprocess.PIPE)
-    py = subprocess.Popen([
-        sys.executable, "-m", "cutadapt", "--cores", str(cores),
-        "-a", "TTAGACATATCTCCGTCG", "-o", out_path, "-"],
-        stdin=cat.stdout)
-    _ = py.communicate()
-    cat.stdout.close()
-    _ = py.communicate()[0]
-    assert_files_equal(cutpath("small.fastq"), out_path)
-
-
-def test_standard_output(tmpdir, cores):
-    """Write FASTQ to standard output (not using --output/-o option)"""
-
-    import subprocess
-    out_path = str(tmpdir.join("out.fastq"))
-    with open(out_path, "w") as out_file:
-        py = subprocess.Popen([
-            sys.executable, "-m", "cutadapt", "--cores", str(cores),
-            "-a", "TTAGACATATCTCCGTCG", datapath("small.fastq")],
-            stdout=out_file)
-        _ = py.communicate()
-    assert_files_equal(cutpath("small.fastq"), out_path)
-
-
-def test_explicit_standard_output(tmpdir, cores):
-    """Write FASTQ to standard output (using "-o -")"""
-
-    out_path = str(tmpdir.join("out.fastq"))
-    with open(out_path, "w") as out_file:
-        py = subprocess.Popen([
-            sys.executable, "-m", "cutadapt", "-o", "-", "--cores", str(cores),
-            "-a", "TTAGACATATCTCCGTCG", datapath("small.fastq")],
-            stdout=out_file)
-        _ = py.communicate()
-    assert_files_equal(cutpath("small.fastq"), out_path)
-
-
-def test_force_fasta_output(tmpdir, cores):
-    """Write FASTA to standard output even on FASTQ input"""
-
-    out_path = str(tmpdir.join("out.fasta"))
-    with open(out_path, "w") as out_file:
-        py = subprocess.Popen([
-            sys.executable, "-m", "cutadapt", "--fasta", "-o", "-", "--cores", str(cores),
-            "-a", "TTAGACATATCTCCGTCG", datapath("small.fastq")],
-            stdout=out_file)
-        _ = py.communicate()
-    assert_files_equal(cutpath("small.fasta"), out_path)
-
-
 def test_empty_read_with_wildcard_in_adapter(run):
     run("-g CWC", "empty.fastq", "empty.fastq")
 
@@ -699,11 +678,13 @@ def test_adapter_order(run):
 
 
 def test_reverse_complement_normalized(run):
-    run(
-        "--revcomp -g ^TTATTTGTCT -g ^TCCGCACTGG",
+    stats = run(
+        "--revcomp --no-index -g ^TTATTTGTCT -g ^TCCGCACTGG",
         "revcomp-single-normalize.fastq",
         "revcomp.1.fastq",
     )
+    assert stats.n == 6
+    assert stats.reverse_complemented == 2
 
 
 def test_reverse_complement_and_info_file(run, tmp_path, cores):
@@ -711,6 +692,7 @@ def test_reverse_complement_and_info_file(run, tmp_path, cores):
     run(
         [
             "--revcomp",
+            "--no-index",
             "-g",
             "^TTATTTGTCT",
             "-g",
@@ -729,7 +711,8 @@ def test_reverse_complement_and_info_file(run, tmp_path, cores):
 
 
 def test_max_expected_errors(run, cores):
-    run("--max-ee=0.9", "maxee.fastq", "maxee.fastq")
+    stats = run("--max-ee=0.9", "maxee.fastq", "maxee.fastq")
+    assert stats.too_many_expected_errors == 2
 
 
 def test_max_expected_errors_fasta(tmp_path):
@@ -741,3 +724,24 @@ def test_max_expected_errors_fasta(tmp_path):
 def test_warn_if_en_dashes_used():
     with pytest.raises(SystemExit):
         main(["–q", "25", "-o", "/dev/null", "in.fastq"])
+
+
+@pytest.mark.parametrize("opt", ["-y", "--suffix"])
+def test_suffix(opt, run):
+    """-y/--suffix parameter"""
+    run([opt, ' {name}', '-e', '0', '-a', 'OnlyT=TTTTTTTT', '-a', 'OnlyG=GGGGGGGG'], "suffix.fastq", "suffix.fastq")
+
+
+@pytest.mark.parametrize("opt", ["--prefix", "--suffix"])
+def test_rename_cannot_be_combined_with_other_renaming_options(opt):
+    with pytest.raises(SystemExit):
+        main([opt, "something", "--rename='{id} {comment} extrainfo'", "-o", "/dev/null", datapath("empty.fastq")])
+
+
+def test_rename(run):
+    run([
+        "--rename={id}_{cut_suffix} {header} {adapter_name}",
+        "--cut=-4",
+        "-a", "OnlyT=TTTTTT",
+        "-a", "OnlyG=GGGGGG",
+    ], "rename.fastq", "suffix.fastq")
