@@ -18,9 +18,10 @@ def run_paired(tmpdir):
         path2 = str(tmpdir.join(expected2))
         params += ["-o", path1, "-p", path2]
         params += [datapath(in1), datapath(in2)]
-        assert main(params) is None
+        stats = main(params)
         assert_files_equal(cutpath(expected1), path1)
         assert_files_equal(cutpath(expected2), path2)
+        return stats
 
     return _run
 
@@ -45,11 +46,12 @@ def run_interleaved(tmpdir):
         if expected2:
             tmp2 = str(tmpdir.join("out2-" + expected2))
             params += ["-p", tmp2]
-            assert main(params + paths) is None
+            stats = main(params + paths)
             assert_files_equal(cutpath(expected2), tmp2)
         else:
-            assert main(params + paths) is None
+            stats = main(params + paths)
         assert_files_equal(cutpath(expected1), tmp1)
+        return stats
 
     return _run
 
@@ -102,13 +104,13 @@ def test_untrimmed_paired_output_automatic_pair_filter(tmpdir, run_paired):
 
 
 def test_explicit_format_with_paired(tmpdir, run_paired):
-    # Use --format=fastq with input files whose extension is .txt
+    # Use FASTQ input files whose extension is .txt
     txt1 = str(tmpdir.join("paired.1.txt"))
     txt2 = str(tmpdir.join("paired.2.txt"))
     shutil.copyfile(datapath("paired.1.fastq"), txt1)
     shutil.copyfile(datapath("paired.2.fastq"), txt2)
     run_paired(
-        "--format=fastq -a TTAGACATAT -m 14 -q 10",
+        "-a TTAGACATAT -m 14 -q 10",
         in1=txt1, in2=txt2,
         expected1="paired.m14.1.fastq",
         expected2="paired.m14.2.fastq",
@@ -410,15 +412,16 @@ def test_nextseq_paired(run_paired, cores):
         cores=cores)
 
 
-def test_paired_demultiplex(tmpdir):
+def test_paired_demultiplex(tmpdir, cores):
     multiout1 = str(tmpdir.join("demultiplexed.{name}.1.fastq"))
     multiout2 = str(tmpdir.join("demultiplexed.{name}.2.fastq"))
     params = [
+        "--cores", str(cores),
         "-a", "first=AACATTAGACA", "-a", "second=CATTAGACATATCGG",
         "-A", "ignored=CAGTGGAGTA", "-A", "alsoignored=AATAACAGTGGAGTA",
         "-o", multiout1, "-p", multiout2,
         datapath("paired.1.fastq"), datapath("paired.2.fastq")]
-    assert main(params) is None
+    main(params)
     assert_files_equal(cutpath("demultiplexed.first.1.fastq"), multiout1.format(name="first"))
     assert_files_equal(cutpath("demultiplexed.second.1.fastq"), multiout1.format(name="second"))
     assert_files_equal(cutpath("demultiplexed.unknown.1.fastq"), multiout1.format(name="unknown"))
@@ -496,13 +499,14 @@ def test_pair_adapters_unequal_length(tmpdir):
         ])
 
 
-def test_pair_adapters_demultiplexing(tmpdir):
+def test_pair_adapters_demultiplexing(tmpdir, cores):
     params = "-g i1=AAAA -G i1=GGGG -g i2=CCCC -G i2=TTTT".split()
     params += ["--pair-adapters"]
+    params += ["--cores", str(cores)]
     params += ["-o", str(tmpdir.join("dual-{name}.1.fastq"))]
     params += ["-p", str(tmpdir.join("dual-{name}.2.fastq"))]
     params += [datapath("dual-index.1.fastq"), datapath("dual-index.2.fastq")]
-    assert main(params) is None
+    main(params)
     for name in [
         "dual-i1.1.fastq",
         "dual-i1.2.fastq",
@@ -516,39 +520,33 @@ def test_pair_adapters_demultiplexing(tmpdir):
 
 
 @pytest.mark.parametrize("discarduntrimmed", (False, True))
-def test_combinatorial_demultiplexing(tmpdir, discarduntrimmed):
+def test_combinatorial_demultiplexing(tmpdir, discarduntrimmed, cores):
     params = "-g A=^AAAAAAAAAA -g C=^CCCCCCCCCC -G G=^GGGGGGGGGG -G T=^TTTTTTTTTT".split()
-    params += ["-o", str(tmpdir.join("combinatorial.{name1}_{name2}.1.fasta"))]
-    params += ["-p", str(tmpdir.join("combinatorial.{name1}_{name2}.2.fasta"))]
-    params += [datapath("combinatorial.1.fasta"), datapath("combinatorial.2.fasta")]
-    combinations = [
-        # third column says whether the file must exist
-        ("A", "G", True),
-        ("A", "T", True),
-        ("C", "G", True),
-        ("C", "T", True),
-    ]
+    params += ["-o", str(tmpdir.join("combinatorial.{name1}_{name2}.1.fastq"))]
+    params += ["-p", str(tmpdir.join("combinatorial.{name1}_{name2}.2.fastq"))]
+    params += ["--cores", str(cores)]
+    params += [datapath("combinatorial.1.fastq"), datapath("combinatorial.2.fastq")]
+    # third item in tuple says whether the file must exist
+    combinations = [(a, b, True) for a, b in product("AC", "GT")]
+    optional = [("unknown", "unknown")]
+    optional += [(a, "unknown") for a in "AC"]
+    optional += [("unknown", b) for b in "GT"]
     if discarduntrimmed:
-        combinations += [
-            ("unknown", "G", False),
-            ("A", "unknown", False),
-        ]
+        combinations.extend((a, b, False) for a, b in optional)
         params += ["--discard-untrimmed"]
     else:
-        combinations += [
-            ("unknown", "G", True),
-            ("A", "unknown", True),
-        ]
-    assert main(params) is None
+        combinations.extend((a, b, True) for a, b in optional)
+    main(params)
     for (name1, name2, should_exist) in combinations:
         for i in (1, 2):
-            name = "combinatorial.{name1}_{name2}.{i}.fasta".format(name1=name1, name2=name2, i=i)
+            name = "combinatorial.{name1}_{name2}.{i}.fastq".format(name1=name1, name2=name2, i=i)
             path = cutpath(os.path.join("combinatorial", name))
             if should_exist:
-                assert tmpdir.join(name).check(), "Output file missing"
-                assert_files_equal(path, str(tmpdir.join(name)))
+                assert tmpdir.join(name).check(), ("Output file missing", name)
+                if os.path.exists(path):
+                    assert_files_equal(path, str(tmpdir.join(name)))
             else:
-                assert not tmpdir.join(name).check(), "Output file should not exist"
+                assert not tmpdir.join(name).check(), ("Output file should not exist", name)
 
 
 def test_info_file(tmpdir):
@@ -560,4 +558,17 @@ def test_info_file(tmpdir):
         datapath("paired.1.fastq"),
         datapath("paired.2.fastq"),
     ]
-    assert main(params) is None
+    main(params)
+
+
+def test_rename(run_paired, cores):
+    run_paired([
+            "--rename={id} {r1.cut_prefix} {cut_prefix}"
+            " {comment} {adapter_name} {r2.adapter_name}",
+            "--cut=4",
+            "-a", "R1adapter=GTCTCCAGCT", "-A", "R2adapter=GACAAATAAC",
+        ],
+        in1="paired.1.fastq", in2="paired.2.fastq",
+        expected1="rename.1.fastq", expected2="rename.2.fastq",
+        cores=cores,
+    )
