@@ -5,7 +5,7 @@ from cutadapt.modifiers import AdapterCutter, ModificationInfo
 
 def test_statistics():
     read = Sequence('name', 'AAAACCCCAAAA')
-    adapters = [BackAdapter("CCCC", max_error_rate=0.1)]
+    adapters = [BackAdapter("CCCC", max_errors=0.1)]
     cutter = AdapterCutter(adapters, times=3)
     cutter(read, ModificationInfo(read))
     # TODO make this a lot simpler
@@ -25,7 +25,7 @@ def test_end_trim_with_mismatch():
     the hit and so the match is considered good. An insertion or substitution
     at the same spot is not a match.
     """
-    adapter = BackAdapter("TCGATCGATCGAT", max_error_rate=0.1)
+    adapter = BackAdapter("TCGATCGATCGAT", max_errors=0.1)
 
     read = Sequence('foo1', 'AAAAAAAAAAATCGTCGATC')
     cutter = AdapterCutter([adapter], times=1)
@@ -46,7 +46,7 @@ def test_end_trim_with_mismatch():
 
 
 def test_anywhere_with_errors():
-    adapter = AnywhereAdapter("CCGCATTTAG", max_error_rate=0.1)
+    adapter = AnywhereAdapter("CCGCATTTAG", max_errors=0.1)
     for seq, expected_trimmed in (
         ('AACCGGTTccgcatttagGATC', 'AACCGGTT'),
         ('AACCGGTTccgcgtttagGATC', 'AACCGGTT'),  # one mismatch
